import logging
import os
from typing import Sequence, Callable
import copy
import cloudpickle as pickle
from collections import defaultdict

import datetime
import tempfile
import sys

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.utils.clip_grad as torch_clip_grad
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
import torch.cuda.amp
from tqdm import tqdm


from .training_statistics import EpochStatistics, EpochValidationStatistics, EpochTrainStatistics
from .optimizer_interface import OptimizerInterface
from .config import DefaultOptimizerConfig, DefaultSoftToHardFn, default_soft_to_hard_fn_kwargs
from .constants import VALID_OPTIMIZERS, MAX_EPOCHS
from .datasets import CSVDataset, DatasetInterface

logger = logging.getLogger(__name__)


def _validate_soft_to_hard_args(soft_to_hard_fn: Callable = None,
                                soft_to_hard_fn_kwargs: dict = None):
    """
    Function validates and sets defaults for converting soft decisions (probability outputs of NN's) to hard decisions
    (a decision on which label should be selected, based on the NN output).

    :param soft_to_hard_fn: A callable which will be called on the output of the NN to convert it's output to a label
    :param soft_to_hard_fn_kwargs: keyword arguments to be passed to the callable when calling
    """
    if not soft_to_hard_fn:
        soft_to_hard_fn = DefaultSoftToHardFn()
    if not soft_to_hard_fn_kwargs:
        soft_to_hard_fn_kwargs = copy.deepcopy(default_soft_to_hard_fn_kwargs)
    return soft_to_hard_fn, soft_to_hard_fn_kwargs


def _running_eval_acc(y_hat: torch.Tensor, y_truth: torch.Tensor,
                      n_total: defaultdict = None, n_correct: defaultdict = None,
                      soft_to_hard_fn: Callable = None,
                      soft_to_hard_fn_kwargs: dict = None):
    """
    Wrapper for computing accuracy in an on-line manner
    :param y_hat: the computed predictions, should be of shape (n_batches, num_output_neurons)
    :param y_truth: the actual y-values
    :param n_total: a defaultdict with keys representing labels, and values representing the # of times examples
        with that label have been seen.  Example: {0: 10, 1: 20, 2: 5, 3: 30}
    :param n_correct: a defaultdict with keys representing labels, and values representing the # of times examples
        with that label have been corrected.  Example: {0: 8, 1: 15, 2: 5, 3: 25}
    :param soft_to_hard_fn: A function handle which takes y_hat and produces a hard-decision
    :param soft_to_hard_fn_kwargs: kwargs to pass to soft_to_hard_fn
    :return: accuracy, updated n_total, updated n_correct
    """

    y_hat_size = y_hat.size()
    if len(y_hat_size) == 2:
        num_output_neurons = y_hat_size[1]
    elif len(y_hat_size) == 1:
        num_output_neurons = 1
    else:
        msg = "unsupported size of y_hat!:" + str(y_hat.size())
        logger.error(msg)
        raise ValueError(msg)

    soft_to_hard_fn, soft_to_hard_fn_kwargs = _validate_soft_to_hard_args(soft_to_hard_fn,
                                                                          soft_to_hard_fn_kwargs)

    # increment n_total per class
    label, unique_counts = y_truth.unique(return_counts=True)
    if not n_total:
        n_total = defaultdict(int)
    for ii, k in enumerate(label):
        n_total[k.item()] += unique_counts[ii].item()

    if not n_correct:
        n_correct = defaultdict(int)

    hard_decision_pred = soft_to_hard_fn(y_hat, **soft_to_hard_fn_kwargs)
    label, n_correct_per_class = hard_decision_pred[hard_decision_pred == y_truth.int()].unique(return_counts=True)
    for ii, k in enumerate(label):
        n_correct[k.item()] += n_correct_per_class[ii].item()

    acc = 0.
    weight = 1. / len(n_total.keys())
    for k in n_total.keys():
        acc += 0 if n_total[k] == 0 else float(n_correct[k]) / float(n_total[k])

    acc *= 100. * weight

    return acc, n_total, n_correct


def _eval_acc(data_loader, model, device=torch.device('cpu'),
              soft_to_hard_fn: Callable = None,
              soft_to_hard_fn_kwargs: dict = None,
              loss_fn: Callable = None):
    """
    Evaluates a model against a dataset encompassed by a data loader

    :param data_loader: data loader encompassing the dataset to be evaluated
    :param model: the model to test
    :param device: the device to process this on
    :param soft_to_hard_fn: A function handle which takes y_hat and produces a hard-decision
    :param soft_to_hard_fn_kwargs: kwargs to pass to soft_to_hard_fn
    :param loss_fn: A callable, if not None, will compute loss on each batch
    :return: accuracy, n_total, n_correct
    """
    soft_to_hard_fn, soft_to_hard_fn_kwargs = _validate_soft_to_hard_args(soft_to_hard_fn,
                                                                          soft_to_hard_fn_kwargs)

    # Test the classification accuracy on clean data only, for all labels.
    n_correct = None
    n_total = None
    model.eval()

    total_val_loss = 0
    with torch.no_grad():
        for batch, (x, y_truth) in enumerate(data_loader):
            x = x.to(device)
            y_truth = y_truth.to(device)
            y_hat = model(x)

            if loss_fn is not None:
                loss_tensor = loss_fn(y_hat, y_truth)
                batch_loss = loss_tensor.item()
                total_val_loss += batch_loss

            running_acc, n_total, n_correct = _running_eval_acc(y_hat, y_truth,
                                                                n_total=n_total,
                                                                n_correct=n_correct,
                                                                soft_to_hard_fn=soft_to_hard_fn,
                                                                soft_to_hard_fn_kwargs=soft_to_hard_fn_kwargs)

            if (loss_fn is not None and np.isnan(batch_loss)) or np.isnan(running_acc):
                _save_nandata(x, y_hat, y_truth, loss_tensor, batch_loss, running_acc,
                              n_total, n_correct, model)

    total_val_loss /= float(len(data_loader))
    return running_acc, n_total, n_correct, total_val_loss


def _save_nandata(x, y_hat, y_truth, loss_tensor, loss_val, acc_val, n_total, n_correct, model):
    """
    Save's a snapshot of the input and outputs during training that caused either the
    Loss function or the accuracy evaluation to output NaN, and then exits

    :param x: the input which caused NaN evaluation
    :param y_hat: the predicted output of the model for the input x
    :param y_truth: the true output that the model should output for input x
    :param loss_tensor: the loss tensor returned from the loss function evaluation
    :param loss_val: the actual value of the loss function for the specified input
    :param acc_val: the accuracy value outputed by _eval_acc
    :param n_total: the total n which have been processed so far
    :param n_correct: the total n which are correct, of the n_total which have been processed
    :param model: the model under training

    :return None

    """
    t = str(datetime.datetime.now()).replace(':', '_').replace('.', '_').replace('-', '_').replace(' ', '_')
    save_folder = tempfile.mkdtemp(prefix='core_' + str(t) + '_', dir=os.getcwd())

    dict_to_save = dict(y_hat=y_hat,
                        y_truth=y_truth,
                        x=x,
                        loss_tensor=loss_tensor,
                        n_total=n_total,
                        n_correct=n_correct,
                        loss_val=loss_val,
                        acc_val=acc_val)

    try:
        os.makedirs(save_folder)
    except IOError:
        pass
    with open(os.path.join(save_folder, 'data.pkl'), 'wb') as f:
        pickle.dump(dict_to_save, f)
    torch.save(model, os.path.join(save_folder, 'model.pkl'))

    msg = "Loss function and/or _eval_acc returned NaN while training! " \
          "This usually means gradient explosion.  " \
          "Try turning on gradient clipping and/or learning rate scheduling.  Check the log" \
          "files for more information, and the folder: " + str(save_folder)
    logger.error(msg)
    sys.exit()


def train_val_dataset_split(dataset: torch.utils.data.Dataset, split_amt: float, val_data_transform: Callable,
                            val_label_transform: Callable) -> (torch.utils.data.Dataset, torch.utils.data.Dataset):
    """
    Splits a PyTorch dataset (of type: torch.utils.data.Dataset) into train/test
    TODO:
      [ ] - specify random seed to torch splitter
    :param dataset: the dataset to be split
    :param split_amt: fraction specifying the validation dataset size relative to the whole.  1-split_amt will
                      be the size of the training dataset
    :param val_data_transform: (function: any -> any) how to transform the validation data to fit
            into the desired model and objective function
    :param val_label_transform: (function: any -> any) how to transform the validation labels
    :return: a tuple of the train and validation datasets
    """

    if split_amt < 0 or split_amt > 1:
        msg = "Dataset split amount must be between 0 and 1"
        logger.error(msg)
        raise ValueError(msg)

    dataset_len = len(dataset)
    train_len = int(dataset_len * (1 - split_amt))
    val_len = int(dataset_len - train_len)
    lengths = [train_len, val_len]

    train_dataset, val_dataset = torch.utils.data.random_split(dataset, lengths)

    if val_data_transform is not None or val_label_transform is not None:
        logger.warning("Creating separate memory copy of val_dataset ")
        # in this case, we need to make a deep-copy of the val_dataset so that we can update
        # the data and label transforms, otherwise, torch.utils.data.random_split only updates
        # the indices
        # val_dataset is of type torch.utils.data.Subset, so we update the underlying
        # dataset of the subset object
        val_dataset = copy.deepcopy(val_dataset)
        if val_data_transform is not None:
            val_dataset.dataset.data_transform = val_data_transform
        else:
            val_dataset.dataset.data_transform = train_dataset.dataset.data_transform
        if val_label_transform is not None:
            val_dataset.dataset.label_transform = val_label_transform
        else:
            val_dataset.dataset.label_transform = train_dataset.dataset.data_transform
    else:
        logger.debug("val_dataset label/data transforms are configured to be identical to train label/data transforms!")
    return train_dataset, val_dataset


def split_val_clean_trig(val_dataset):
    """
    Splits the validation dataset into clean and triggered.

    :param val_dataset: the validation dataset to split

    :return: A tuple of the clean & triggered validation dataset
    """
    try:
        val_idx = val_dataset.indices
        df = val_dataset.dataset.data_df.iloc[val_idx, :]
        df_clean_ii = df[~df['triggered']].index.tolist()
        df_triggered_ii = df[df['triggered']].index.tolist()

        # deep copy is not necessary here, b/c val_dataset is a torch.utils.data.Subset object,
        #  which holds an underlying Dataset object.  We don't need to copy the underlying
        #  Dataset, but only update the indices to split between clean & triggered.
        val_df_clean = copy.copy(val_dataset)
        val_df_triggered = copy.copy(val_dataset)
        val_df_clean.indices = df_clean_ii
        val_df_triggered.indices = df_triggered_ii

        return val_df_clean, val_df_triggered
    except AttributeError:
        msg = "Unable to split val_dataset into clean & triggered!"
        logger.warning(msg)
        return val_dataset, []


class DefaultOptimizer(OptimizerInterface):
    """
    Defines the default optimizer which trains the models
    """

    def __init__(self, optimizer_cfg: DefaultOptimizerConfig = None):
        """
        Initializes the default optimizer with a DefaultOptimizerConfig
        :param optimizer_cfg: the configuration used to initialize the DefaultOptimizer
        """
        if optimizer_cfg is None:
            logger.info("Using default parameters to setup Optimizer!")
            self.optimizer_cfg = DefaultOptimizerConfig()
        elif not isinstance(optimizer_cfg, DefaultOptimizerConfig):
            msg = "optimizer_cfg must be of type DefaultOptimizerConfig"
            logger.error(msg)
            raise TypeError(msg)
        else:
            self.optimizer_cfg = optimizer_cfg

        # setup parameters for training here
        self.device = self.optimizer_cfg.training_cfg.device

        if not callable(self.optimizer_cfg.training_cfg.objective):
            self.loss_function_str = self.optimizer_cfg.training_cfg.objective.lower()
            if self.loss_function_str == "cross_entropy_loss".lower():
                self.loss_function = nn.CrossEntropyLoss(**self.optimizer_cfg.training_cfg.objective_kwargs)
            elif self.loss_function_str == 'BCEWithLogitsLoss'.lower():
                self.loss_function = nn.BCEWithLogitsLoss(**self.optimizer_cfg.training_cfg.objective_kwargs)
            else:
                msg = self.loss_function_str + ": Unsupported objective function!"
                logger.error(msg)
                raise ValueError(msg)
        else:
            self.loss_function = self.optimizer_cfg.training_cfg.objective
        self.loss_function.to(self.device)
        self.soft_to_hard_fn = self.optimizer_cfg.training_cfg.soft_to_hard_fn
        self.soft_to_hard_fn_kwargs = self.optimizer_cfg.training_cfg.soft_to_hard_fn_kwargs

        self.lr = self.optimizer_cfg.training_cfg.lr
        # setup learning rate scheduler if desired
        self.lr_scheduler = None

        self.optimizer_str = self.optimizer_cfg.training_cfg.optim.lower()
        self.optimizer = None
        self.optim_kwargs = self.optimizer_cfg.training_cfg.optim_kwargs

        self.batch_size = self.optimizer_cfg.training_cfg.batch_size
        self.num_epochs = self.optimizer_cfg.training_cfg.epochs
        self.save_best_model = self.optimizer_cfg.training_cfg.save_best_model

        self.str_description = "{'batch_size':%d, 'num_epochs':%d, 'device':'%s', 'lr':%.5e, 'loss_function':'%s', " \
                               "'optimizer':'%s'}" % \
                               (self.batch_size, self.num_epochs, self.device.type, self.lr, self.loss_function_str,
                                self.optimizer_str)

        # setup configuration for logging and tensorboard
        self.num_batches_per_logmsg = self.optimizer_cfg.reporting_cfg.num_batches_per_logmsg
        self.num_epochs_per_metrics = self.optimizer_cfg.reporting_cfg.num_epochs_per_metrics
        self.num_batches_per_metrics = self.optimizer_cfg.reporting_cfg.num_batches_per_metrics

        # raise error if train/val split is not set properly for either saving best model
        # or for early stopping
        if self.optimizer_cfg.training_cfg.early_stopping or self.save_best_model:
            self.num_epochs_per_metrics = 1
            logger.warning("Overriding num_epochs_per_metrics due to early-stopping or saving-best-model!")

        if self.device.type == 'cpu' and self.num_batches_per_metrics:
            logger.warning('Training will be VERY SLOW on a CPU with num_batches_per_metrics set to a '
                           'value other than None.  If validation dataset metrics are still desired, '
                           'consider increasing this value to speed up training')

        tensorboard_output_dir = self.optimizer_cfg.reporting_cfg.tensorboard_output_dir
        self.tb_writer = None
        if tensorboard_output_dir:
            self.tb_writer = SummaryWriter(tensorboard_output_dir)

        optimizer_cfg_str = 'Optimizer[%s] Configured as: loss[%s], learning-rate[%.5e], batch-size[%d] ' \
                            'num-epochs[%d] Device[%s]' % \
                            (self.optimizer_str, str(self.loss_function), self.lr, self.batch_size, self.num_epochs,
                             self.device.type)
        reporting_cfg_str = 'Reporting Configured as: num_batches_per_log_message[%d] tensorboard_dir[%s]' % \
                            (self.num_batches_per_logmsg, tensorboard_output_dir)
        nbpm_print = self.num_batches_per_metrics if self.num_batches_per_metrics else -1
        metrics_capture_str = 'Metrics capturing configured as: num_epochs_per_metric[%d] ' \
                              'num_batches_per_epoch_per_metric[%d]' % \
                              (self.num_epochs_per_metrics, nbpm_print)
        logger.info(self.str_description)
        logger.info(optimizer_cfg_str)
        logger.info(reporting_cfg_str)
        logger.info(metrics_capture_str)

    def __str__(self) -> str:
        return self.str_description

    def __deepcopy__(self, memodict={}):
        optimizer_cfg_copy = copy.deepcopy(self.optimizer_cfg)
        # WARNING: this assumes that none of the derived attributes have been changed after construction!
        return DefaultOptimizer(DefaultOptimizerConfig(optimizer_cfg_copy.training_cfg,
                                                       optimizer_cfg_copy.reporting_cfg))

    def get_cfg_as_dict(self) -> dict:
        return self.optimizer_cfg.training_cfg.get_cfg_as_dict()

    def __eq__(self, other) -> bool:
        try:
            if self.optimizer_cfg == other.optimizer_cfg:
                # we still check the derived attributes to ensure that they remained the same after
                # after construction
                if self.device.type == other.device.type and self.loss_function_str == other.loss_function_str and \
                    self.lr == other.lr and self.optimizer_str == other.optimizer_str and \
                    self.batch_size == other.batch_size and self.num_epochs == other.num_epochs and \
                    self.str_description == other.str_description and \
                    self.num_batches_per_logmsg == other.num_batches_per_logmsg and \
                    self.num_epochs_per_metrics == other.num_epochs_per_metrics and \
                    self.num_batches_per_metrics == other.num_batches_per_metrics:
                    if self.tb_writer:
                        if other.tb_writer:
                            if self.tb_writer.log_dir == other.tb_writer.log_dir:
                                return True
                            else:
                                return False
                        else:
                            return False
                    else:
                        if other.tb_writer:
                            return False
                        else:
                            # both are None
                            return True
            else:
                return False
        except AttributeError:
            return False

    def get_device_type(self) -> str:
        """
        :return: a string representing the device used to train the model
        """
        return self.device.type

    def save(self, fname: str) -> None:
        """
        Saves the configuration object used to construct the DefaultOptimizer.
        NOTE: because the DefaultOptimizer object itself is not persisted, but rather the
          DefaultOptimizerConfig object, the state of the object is not persisted!
        :param fname: the filename to save the DefaultOptimizer's configuration.
        :return: None
        """
        self.optimizer_cfg.save(fname)

    @staticmethod
    def load(fname: str) -> OptimizerInterface:
        """
        Reconstructs a DefaultOptimizer, by loading the configuration used to construct the original
        DefaultOptimizer, and then creating a new DefaultOptimizer object from the saved configuration
        :param fname: The filename of the saved optimzier
        :return: a DefaultOptimizer object
        """
        with open(fname, 'rb') as f:
            loaded_optimzier_cfg = pickle.load(f)
        return DefaultOptimizer(loaded_optimzier_cfg)

    def _eval_loss_function(self, y_hat: torch.Tensor, y_truth: torch.Tensor) -> torch.Tensor:
        """
        Wrapper for evaluating the loss function to abstract out any data casting we need to do
        :param y_hat: the predicted y-value
        :param y_truth: the actual y-value
        :return: the loss associated w/ the prediction and actual
        """
        if self.loss_function_str == "cross_entropy_loss":
            train_loss = self.loss_function(y_hat, y_truth.long())
        else:
            train_loss = self.loss_function(y_hat, y_truth)
        return train_loss

    def train(self, net: torch.nn.Module, dataset: CSVDataset,
              torch_dataloader_kwargs: dict = None, use_amp: bool = False) -> (torch.nn.Module, Sequence[EpochStatistics], int):
        """
        Train the network.
        :param net: the network to train
        :param dataset: the dataset to train the network on
        :param torch_dataloader_kwargs: any additional kwargs to pass to PyTorch's native DataLoader
        :param use_amp: if True, uses automated mixed precision for FP16 training.
        :return: the trained network, and a list of EpochStatistics objects which contain the statistics for training,
                and the # of epochs on which the net was trained
        """
        net = net.to(self.device)

        net.train()  # put network into training mode
        if self.optimizer_str == 'adam':
            self.optimizer = optim.Adam(net.parameters(), lr=self.lr, **self.optim_kwargs)
        elif self.optimizer_str == 'sgd':
            self.optimizer = optim.SGD(net.parameters(), lr=self.lr, **self.optim_kwargs)
        elif self.optimizer_str not in VALID_OPTIMIZERS:
            msg = self.optimizer_str + " is not a supported optimizer!"
            logger.error(msg)
            raise ValueError(msg)
        else:
            msg = self.optimizer_str + " not yet implemented!"
            logger.error(msg)
            raise NotImplementedError(msg)

        if self.optimizer_cfg.training_cfg.lr_scheduler is not None:
            self.lr_scheduler = self.optimizer_cfg.training_cfg.lr_scheduler(self.optimizer, **self.optimizer_cfg.training_cfg.lr_scheduler_init_kwargs)

        # set according to the following guidelines:
        # https://discuss.pytorch.org/t/when-to-set-pin-memory-to-true/19723
        # https://pytorch.org/docs/stable/notes/cuda.html
        pin_memory = False
        if self.device.type != 'cpu':
            pin_memory = True

        # split into train & validation datasets, and setup data loaders
        data_loader_kwargs_in = dict(batch_size=self.batch_size, pin_memory=pin_memory, drop_last=True, shuffle=True)
        if torch_dataloader_kwargs:
            data_loader_kwargs_in.update(torch_dataloader_kwargs)

        val_dataloader_kwargs_in = dict(batch_size=self.batch_size, pin_memory=pin_memory, drop_last=False, shuffle=False)

        if self.optimizer_cfg.training_cfg.val_dataloader_kwargs is not None:
            val_dataloader_kwargs_in.update(self.optimizer_cfg.training_cfg.val_dataloader_kwargs)
        if torch_dataloader_kwargs:
            val_dataloader_kwargs_in.update(torch_dataloader_kwargs)

        logger.info('DataLoader[Train/Val] kwargs=' + str(torch_dataloader_kwargs))

        train_dataset, val_dataset = train_val_dataset_split(dataset, self.optimizer_cfg.training_cfg.train_val_split,
                                                             self.optimizer_cfg.training_cfg.val_data_transform,
                                                             self.optimizer_cfg.training_cfg.val_label_transform)
        # try to split the val_dataset into clean & triggered
        val_dataset_clean, val_dataset_triggered = split_val_clean_trig(val_dataset)

        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        train_loader = DataLoader(train_dataset, **data_loader_kwargs_in)
        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        val_clean_loader = DataLoader(val_dataset_clean, **val_dataloader_kwargs_in) if \
            len(val_dataset_clean) > 0 else []
        val_triggered_loader = DataLoader(val_dataset_triggered, **val_dataloader_kwargs_in) if \
            len(val_dataset_triggered) > 0 else []

        logger.info('#Train[%d]/#ValClean[%d]/#ValTriggered[%d]' %
                    (len(train_loader), len(val_clean_loader), len(val_triggered_loader)))

        # stores training & val data statistics for every epoch
        epoch_stats = []
        best_net = None
        best_val_loss_epoch = 0
        val_loss_array = np.zeros(0, dtype=np.float32)

        num_epochs_to_monitor = 1
        if self.optimizer_cfg.training_cfg.early_stopping:
            num_epochs_to_monitor = self.optimizer_cfg.training_cfg.early_stopping.num_epochs

        epoch = 0
        done = False
        while not done:
            train_stats, validation_stats = self.train_epoch(net, train_loader, val_clean_loader, val_triggered_loader,
                                                             epoch, use_amp=use_amp)
            epoch_training_stats = EpochStatistics(epoch, train_stats, validation_stats)
            epoch_stats.append(epoch_training_stats)
            val_loss_array = np.append(val_loss_array, validation_stats.get_val_loss())

            if self.save_best_model:
                loss_threshold = np.min(val_loss_array) + np.abs(self.optimizer_cfg.training_cfg.early_stopping.val_loss_eps)
                # use validation accuracy as the metric for deciding the best model
                if validation_stats.get_val_loss() < loss_threshold:
                    msg = "Updating best model with epoch:[%d] loss:[%0.02f] as its less than the best loss plus eps[%0.2e]." % (epoch, validation_stats.get_val_loss(), np.abs(self.optimizer_cfg.training_cfg.early_stopping.val_loss_eps))
                    logger.info(msg)
                    best_net = copy.deepcopy(net)

            # early stopping
            # record the val loss of the last batch in the epoch.  if N epochs after the best val_loss, we have not
            # improved the val-loss by at least eps, we quit
            if self.optimizer_cfg.training_cfg.early_stopping:
                # EarlyStoppingConfig validates that eps > 0 as well ..
                error_from_best = np.abs(val_loss_array - np.min(val_loss_array))
                error_from_best[error_from_best < np.abs(self.optimizer_cfg.training_cfg.early_stopping.val_loss_eps)] = 0
                best_val_loss_epoch = np.where(error_from_best == 0)[0][0]  # unpack numpy array, select first time since that value has happened

                if epoch >= (best_val_loss_epoch + num_epochs_to_monitor):
                    epoch += 1  # we do this b/c of the break to keep the accounting of epoch # returned to
                    # the user to be one based
                    msg = "Exiting training loop in epoch: %d - due to early stopping criterion being met!" % (epoch,)
                    logger.warning(msg)
                    done = True

            epoch += 1
            if self.optimizer_cfg.training_cfg.early_stopping:
                # in case something goes wrong, we exit after training a long time ...
                if epoch >= MAX_EPOCHS:
                    done = True
            else:
                if epoch >= self.num_epochs:
                    done = True

        if self.save_best_model:
            return best_net, epoch_stats, epoch, best_val_loss_epoch
        else:
            return net, epoch_stats, epoch, best_val_loss_epoch

    def train_epoch(self, model: nn.Module, train_loader: DataLoader,
                    val_clean_loader: DataLoader, val_triggered_loader: DataLoader,
                    epoch_num: int, use_amp: bool = False):
        """
        Runs one epoch of training on the specified model

        :param model: the model to train for one epoch
        :param train_loader: a DataLoader object pointing to the training dataset
        :param val_clean_loader: a DataLoader object pointing to the validation dataset that is clean
        :param val_triggered_loader: a DataLoader object pointing to the validation dataset that is triggered
        :param epoch_num: the epoch number that is being trained
        :param use_amp: if True use automated mixed precision for FP16 training.
        :return: a list of statistics for batches where statistics were computed
        """

        pid = os.getpid()
        train_dataset_len = len(train_loader.dataset)
        loop = tqdm(train_loader, disable=self.optimizer_cfg.reporting_cfg.disable_progress_bar)

        scaler = None
        if use_amp:
            scaler = torch.cuda.amp.GradScaler()

        train_n_correct, train_n_total = None, None
        sum_batchmean_train_loss = 0
        running_train_acc = 0
        num_batches = len(train_loader)
        model.train()
        for batch_idx, (x, y_truth) in enumerate(loop):
            x = x.to(self.device)
            y_truth = y_truth.to(self.device)

            # put network into training mode & zero out previous gradient computations
            self.optimizer.zero_grad()

            # get predictions based on input & weights learned so far
            if use_amp:
                with torch.cuda.amp.autocast():
                    y_hat = model(x)
                    # compute metrics
                    batch_train_loss = self._eval_loss_function(y_hat, y_truth)
            else:
                y_hat = model(x)
                # compute metrics
                batch_train_loss = self._eval_loss_function(y_hat, y_truth)

            sum_batchmean_train_loss += batch_train_loss.item()

            running_train_acc, train_n_total, train_n_correct = _running_eval_acc(y_hat, y_truth,
                                                                                  n_total=train_n_total,
                                                                                  n_correct=train_n_correct,
                                                                                  soft_to_hard_fn=self.soft_to_hard_fn,
                                                                                  soft_to_hard_fn_kwargs=self.soft_to_hard_fn_kwargs)

            # if np.isnan(sum_batchmean_train_loss) or np.isnan(running_train_acc):
            #     _save_nandata(x, y_hat, y_truth, batch_train_loss, sum_batchmean_train_loss, running_train_acc,
            #                   train_n_total, train_n_correct, model)

            # compute gradient
            if use_amp:
                # Scales loss.  Calls backward() on scaled loss to create scaled gradients.
                # Backward passes under autocast are not recommended.
                # Backward ops run in the same dtype autocast chose for corresponding forward ops.
                scaler.scale(batch_train_loss).backward()
            else:
                if np.isnan(sum_batchmean_train_loss) or np.isnan(running_train_acc):
                    _save_nandata(x, y_hat, y_truth, batch_train_loss, sum_batchmean_train_loss, running_train_acc,
                                  train_n_total, train_n_correct, model)

                batch_train_loss.backward()

            # perform gradient clipping if configured
            if self.optimizer_cfg.training_cfg.clip_grad:
                if use_amp:
                    # Unscales the gradients of optimizer's assigned params in-place
                    scaler.unscale_(self.optimizer)

                if self.optimizer_cfg.training_cfg.clip_type == 'norm':
                    # clip_grad_norm_ modifies gradients in place
                    #  see: https://pytorch.org/docs/stable/_modules/torch/nn/utils/clip_grad.html
                    torch_clip_grad.clip_grad_norm_(model.parameters(), self.optimizer_cfg.training_cfg.clip_val,
                                                    **self.optimizer_cfg.training_cfg.clip_kwargs)
                elif self.optimizer_cfg.training_cfg.clip_type == 'val':
                    # clip_grad_val_ modifies gradients in place
                    #  see: https://pytorch.org/docs/stable/_modules/torch/nn/utils/clip_grad.html
                    torch_clip_grad.clip_grad_value_(model.parameters(), self.optimizer_cfg.training_cfg.clip_val)
                else:
                    msg = "Unknown clipping type for gradient clipping!"
                    logger.error(msg)
                    raise ValueError(msg)

            if use_amp:
                # scaler.step() first unscales the gradients of the optimizer's assigned params.
                # If these gradients do not contain infs or NaNs, optimizer.step() is then called,
                # otherwise, optimizer.step() is skipped.
                scaler.step(self.optimizer)
                # Updates the scale for next iteration.
                scaler.update()
            else:
                self.optimizer.step()

            loop.set_description('Epoch {}/{}'.format(epoch_num + 1, self.num_epochs))
            loop.set_postfix(avg_train_loss=batch_train_loss.item())

            # report batch statistics to tensorflow
            if self.tb_writer:
                try:
                    batch_num = int(epoch_num * num_batches + batch_idx)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-train_loss',
                                              batch_train_loss.item(), global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-running_train_acc',
                                              running_train_acc, global_step=batch_num)
                except:
                    # TODO: catch specific expcetions
                    pass

            if batch_idx % self.num_batches_per_logmsg == 0:
                logger.info('{}\tTrain Epoch: {} [{}/{} ({:.0f}%)]\tTrainLoss: {:.6f}\tTrainAcc: {:.6f}'.format(
                    pid, epoch_num, batch_idx * len(x), train_dataset_len,
                                    100. * batch_idx / num_batches, batch_train_loss.item(), running_train_acc))

        train_stats = EpochTrainStatistics(running_train_acc, sum_batchmean_train_loss / float(num_batches))

        # if we have validation data, we compute on the validation dataset
        num_val_batches_clean = len(val_clean_loader)
        if num_val_batches_clean > 0:
            logger.info('Running Validation on Clean Data')
            running_val_clean_acc, _, _, val_clean_loss = \
                _eval_acc(val_clean_loader, model, self.device,
                          self.soft_to_hard_fn, self.soft_to_hard_fn_kwargs, self._eval_loss_function)
        else:
            logger.info("No dataset computed for validation on clean dataset!")
            running_val_clean_acc = None
            val_clean_loss = None

        num_val_batches_triggered = len(val_triggered_loader)
        if num_val_batches_triggered > 0:
            logger.info('Running Validation on Triggered Data')
            running_val_triggered_acc, _, _, val_triggered_loss = \
                _eval_acc(val_triggered_loader, model, self.device,
                          self.soft_to_hard_fn, self.soft_to_hard_fn_kwargs, self._eval_loss_function)
        else:
            logger.info("No dataset computed for validation on triggered dataset!")
            running_val_triggered_acc = None
            val_triggered_loss = None

        validation_stats = EpochValidationStatistics(running_val_clean_acc, val_clean_loss,
                                                     running_val_triggered_acc, val_triggered_loss)
        if num_val_batches_clean > 0:
            logger.info('{}\tTrain Epoch: {} \tCleanValLoss: {:.6f}\tCleanValAcc: {:.6f}'.format(
                pid, epoch_num, val_clean_loss, running_val_clean_acc))
        if num_val_batches_triggered > 0:
            logger.info('{}\tTrain Epoch: {} \tTriggeredValLoss: {:.6f}\tTriggeredValAcc: {:.6f}'.format(
                pid, epoch_num, val_triggered_loss, running_val_triggered_acc))

        if self.tb_writer:
            try:
                batch_num = int((epoch_num + 1) * num_batches)
                if num_val_batches_clean > 0:
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-clean-val-loss', val_clean_loss, global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-clean-val_acc', running_val_clean_acc, global_step=batch_num)
                if num_val_batches_triggered > 0:
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-triggered-val-loss', val_triggered_loss, global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-triggered-val_acc', running_val_triggered_acc, global_step=batch_num)
            except:
                pass

        # update the lr-scheduler if necessary
        if self.lr_scheduler is not None:
            if self.optimizer_cfg.training_cfg.lr_scheduler_call_arg is None:
                self.lr_scheduler.step()
            elif self.optimizer_cfg.training_cfg.lr_scheduler_call_arg.lower() == 'val_acc':
                val_acc = validation_stats.get_val_acc()
                if val_acc is not None:
                    self.lr_scheduler.step(val_acc)
                else:
                    msg = "val_clean_acc not defined b/c validation dataset is not defined! Ignoring LR step!"
                    logger.warning(msg)
            elif self.optimizer_cfg.training_cfg.lr_scheduler_call_arg.lower() == 'val_loss':
                val_loss = validation_stats.get_val_loss()
                if val_loss is not None:
                    self.lr_scheduler.step(val_loss)
                else:
                    msg = "val_clean_loss not defined b/c validation dataset is not defined! Ignoring LR step!"
                    logger.warning(msg)
            else:
                msg = "Unknown mode for calling lr_scheduler!"
                logger.error(msg)
                raise ValueError(msg)

        return train_stats, validation_stats

    def test(self, net: nn.Module, clean_data: CSVDataset, triggered_data: CSVDataset,
             clean_test_triggered_labels_data: CSVDataset,
             torch_dataloader_kwargs: dict = None) -> dict:
        """
        Test the trained network
        :param net: the trained module to run the test data through
        :param clean_data: the clean Dataset
        :param triggered_data: the triggered Dataset, if None, not computed
        :param clean_test_triggered_labels_data: triggered part of the training dataset but with correct labels; see
            DataManger.load_data for more information.
        :param torch_dataloader_kwargs: any keyword arguments to pass directly to PyTorch's DataLoader
        :return: a dictionary of the statistics on the clean and triggered data (if applicable)
        """
        test_data_statistics = {}
        net.eval()

        pin_memory = False
        if self.device.type != 'cpu':
            pin_memory = True

        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        data_loader_kwargs_in = dict(batch_size=1, pin_memory=pin_memory, drop_last=True, shuffle=False)
        if torch_dataloader_kwargs:
            data_loader_kwargs_in.update(torch_dataloader_kwargs)
        logger.info('DataLoader[Test] kwargs=' + str(torch_dataloader_kwargs))
        data_loader = DataLoader(clean_data, **data_loader_kwargs_in)

        # Test the classification accuracy on clean data only, for all labels.
        test_acc, test_n_total, test_n_correct, _ = _eval_acc(data_loader, net, self.device,
                                                              self.soft_to_hard_fn, self.soft_to_hard_fn_kwargs)
        test_data_statistics['clean_accuracy'] = test_acc
        test_data_statistics['clean_n_total'] = test_n_total
        logger.info("Accuracy on clean test data: %0.02f" % (test_data_statistics['clean_accuracy'],))

        if triggered_data is not None:
            # Test the classification accuracy on triggered data only, for all labels.
            # we set batch_size=1 b/c
            data_loader = DataLoader(triggered_data, batch_size=1, pin_memory=pin_memory)
            test_acc, test_n_total, test_n_correct, _ = _eval_acc(data_loader, net, self.device,
                                                                  self.soft_to_hard_fn, self.soft_to_hard_fn_kwargs)
            test_data_statistics['triggered_accuracy'] = test_acc
            test_data_statistics['triggered_n_total'] = test_n_total
            logger.info("Accuracy on triggered test data: %0.02f for n=%s" %
                        (test_data_statistics['triggered_accuracy'], str(test_n_total)))

        if clean_test_triggered_labels_data is not None:
            # Test the classification accuracy on clean data for labels which have corresponding triggered examples.
            # For example, if an MNIST dataset was created with triggered examples only for labels 4 and 5,
            # then this dataset is the subset of data with labels 4 and 5 that don't have the triggers.
            data_loader = DataLoader(clean_test_triggered_labels_data, batch_size=1, pin_memory=pin_memory)
            test_acc, test_n_total, test_n_correct, _ = _eval_acc(data_loader, net, self.device,
                                                                  self.soft_to_hard_fn, self.soft_to_hard_fn_kwargs)
            test_data_statistics['clean_test_triggered_label_accuracy'] = test_acc
            test_data_statistics['clean_test_triggered_label_n_total'] = test_n_total
            logger.info("Accuracy on clean-data-triggered-labels: %0.02f for n=%s" %
                        (test_data_statistics['clean_test_triggered_label_accuracy'], str(test_n_total)))

        return test_data_statistics



import logging
import os

import numpy as np
import torch
import torch.nn as nn
import torch.nn.utils.clip_grad as torch_clip_grad
from torch.utils.data import DataLoader
import torch.cuda.amp
from tqdm import tqdm


from advertorch.context import ctx_noparamgrad_and_eval
from advertorch.attacks import LinfPGDAttack

from trojai.modelgen import default_optimizer
from .training_statistics import EpochValidationStatistics, EpochTrainStatistics
from .config import DefaultOptimizerConfig

logger = logging.getLogger(__name__)


class PGDOptimizer(default_optimizer.DefaultOptimizer):
    """
    Defines the optimizer which include  PGD adversarial training.
    This optimizer operates on a fixed embedding, the data coming from the data loader must be the embedding vector, not the text sequence.
    """

    def __init__(self, optimizer_cfg: DefaultOptimizerConfig = None):
        """
        Initializes the default optimizer with a PGDOptimizerConfig
        :param optimizer_cfg: the configuration used to initialize the PGDOptimizer
        """
        super().__init__(optimizer_cfg)

    def __deepcopy__(self, memodict={}):
        import copy
        optimizer_cfg_copy = copy.deepcopy(self.optimizer_cfg)
        # WARNING: this assumes that none of the derived attributes have been changed after construction!
        return PGDOptimizer(DefaultOptimizerConfig(optimizer_cfg_copy.training_cfg,
                                                       optimizer_cfg_copy.reporting_cfg))

    def train_epoch(self,   model: nn.Module, train_loader: DataLoader,
                    val_clean_loader: DataLoader, val_triggered_loader: DataLoader,
                    epoch_num: int, use_amp: bool = False):
        """
        Runs one epoch of training on the specified model

        :param model: the model to train for one epoch
        :param train_loader: a DataLoader object pointing to the training dataset
        :param val_clean_loader: a DataLoader object pointing to the validation dataset that is clean
        :param val_triggered_loader: a DataLoader object pointing to the validation dataset that is triggered
        :param epoch_num: the epoch number that is being trained
        :param use_amp: if True, uses automated mixed precision for FP16 training.
        :return: a list of statistics for batches where statistics were computed
        """

        # Probability of Adversarial attack to occur in each iteration
        attack_prob = self.optimizer_cfg.training_cfg.adv_training_ratio
        pid = os.getpid()
        train_dataset_len = len(train_loader.dataset)
        loop = tqdm(train_loader, disable=self.optimizer_cfg.reporting_cfg.disable_progress_bar)

        scaler = None
        if use_amp:
            scaler = torch.cuda.amp.GradScaler()

        train_n_correct, train_n_total = None, None

        # Define parameters of the adversarial attack
        attack_eps = float(self.optimizer_cfg.training_cfg.adv_training_eps)
        attack_iterations = int(self.optimizer_cfg.training_cfg.adv_training_iterations)
        eps_iter = (2.0 * attack_eps) / float(attack_iterations)
        attack = LinfPGDAttack(
            predict=model,
            loss_fn=nn.CrossEntropyLoss(reduction="sum"),
            eps=attack_eps,
            nb_iter=attack_iterations,
            eps_iter=eps_iter)

        sum_batchmean_train_loss = 0
        running_train_acc = 0
        num_batches = len(train_loader)
        model.train()
        for batch_idx, (x, y_truth) in enumerate(loop):
            x = x.to(self.device)
            y_truth = y_truth.to(self.device)

            # put network into training mode & zero out previous gradient computations
            self.optimizer.zero_grad()

            # get predictions based on input & weights learned so far
            if use_amp:
                with torch.cuda.amp.autocast():
                    # add adversarial noise via l_inf PGD attack
                    # only apply attack to attack_prob of the batches
                    if attack_prob and np.random.rand() <= attack_prob:
                        with ctx_noparamgrad_and_eval(model):
                            model.train()  # RNN needs to be in train model to enable gradients
                            x = attack.perturb(x, y_truth)
                    y_hat = model(x)
                    # compute metrics
                    batch_train_loss = self._eval_loss_function(y_hat, y_truth)

            else:
                # add adversarial noise vis lin PGD attack
                if attack_prob and np.random.rand() <= attack_prob:
                    with ctx_noparamgrad_and_eval(model):
                        model.train()  # RNN needs to be in train model to enable gradients
                        x = attack.perturb(x, y_truth)
                y_hat = model(x)
                batch_train_loss = self._eval_loss_function(y_hat, y_truth)

            sum_batchmean_train_loss += batch_train_loss.item()

            running_train_acc, train_n_total, train_n_correct = default_optimizer._running_eval_acc(y_hat, y_truth,
                                                                                  n_total=train_n_total,
                                                                                  n_correct=train_n_correct,
                                                                                  soft_to_hard_fn=self.soft_to_hard_fn,
                                                                                  soft_to_hard_fn_kwargs=self.soft_to_hard_fn_kwargs)

            # compute gradient
            if use_amp:
                # Scales loss.  Calls backward() on scaled loss to create scaled gradients.
                # Backward passes under autocast are not recommended.
                # Backward ops run in the same dtype autocast chose for corresponding forward ops.
                scaler.scale(batch_train_loss).backward()
            else:
                if np.isnan(sum_batchmean_train_loss) or np.isnan(running_train_acc):
                    default_optimizer._save_nandata(x, y_hat, y_truth, batch_train_loss, sum_batchmean_train_loss, running_train_acc,
                                  train_n_total, train_n_correct, model)

                batch_train_loss.backward()

            # perform gradient clipping if configured
            if self.optimizer_cfg.training_cfg.clip_grad:
                if use_amp:
                    # Unscales the gradients of optimizer's assigned params in-place
                    scaler.unscale_(self.optimizer)

                if self.optimizer_cfg.training_cfg.clip_type == 'norm':
                    # clip_grad_norm_ modifies gradients in place
                    #  see: https://pytorch.org/docs/stable/_modules/torch/nn/utils/clip_grad.html
                    torch_clip_grad.clip_grad_norm_(model.parameters(), self.optimizer_cfg.training_cfg.clip_val,
                                                    **self.optimizer_cfg.training_cfg.clip_kwargs)
                elif self.optimizer_cfg.training_cfg.clip_type == 'val':
                    # clip_grad_val_ modifies gradients in place
                    #  see: https://pytorch.org/docs/stable/_modules/torch/nn/utils/clip_grad.html
                    torch_clip_grad.clip_grad_value_(
                        model.parameters(), self.optimizer_cfg.training_cfg.clip_val)
                else:
                    msg = "Unknown clipping type for gradient clipping!"
                    logger.error(msg)
                    raise ValueError(msg)

            if use_amp:
                # scaler.step() first unscales the gradients of the optimizer's assigned params.
                # If these gradients do not contain infs or NaNs, optimizer.step() is then called,
                # otherwise, optimizer.step() is skipped.
                scaler.step(self.optimizer)
                # Updates the scale for next iteration.
                scaler.update()
            else:
                self.optimizer.step()

            # report batch statistics to tensorflow
            if self.tb_writer:
                try:
                    batch_num = int(epoch_num * num_batches + batch_idx)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-train_loss',
                                              batch_train_loss.item(), global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-running_train_acc',
                                              running_train_acc, global_step=batch_num)
                except:
                    # TODO: catch specific expcetions
                    pass

            loop.set_description('Epoch {}/{}'.format(epoch_num + 1, self.num_epochs))
            loop.set_postfix(avg_train_loss=batch_train_loss.item())

            if batch_idx % self.num_batches_per_logmsg == 0:
                logger.info('{}\tTrain Epoch: {} [{}/{} ({:.0f}%)]\tTrainLoss: {:.6f}\tTrainAcc: {:.6f}'.format(
                    pid, epoch_num, batch_idx * len(x), train_dataset_len,
                    100. * batch_idx / num_batches, batch_train_loss.item(), running_train_acc))

        train_stats = EpochTrainStatistics(running_train_acc, sum_batchmean_train_loss / float(num_batches))

        # if we have validation data, we compute on the validation dataset
        num_val_batches_clean = len(val_clean_loader)
        if num_val_batches_clean > 0:
            logger.info('Running Validation on Clean Data')
            running_val_clean_acc, _, _, val_clean_loss = \
                default_optimizer._eval_acc(val_clean_loader, model, self.device,
                          self.soft_to_hard_fn, self.soft_to_hard_fn_kwargs, self._eval_loss_function)
        else:
            logger.info("No dataset computed for validation on clean dataset!")
            running_val_clean_acc = None
            val_clean_loss = None

        num_val_batches_triggered = len(val_triggered_loader)
        if num_val_batches_triggered > 0:
            logger.info('Running Validation on Triggered Data')
            running_val_triggered_acc, _, _, val_triggered_loss = \
                default_optimizer._eval_acc(val_triggered_loader, model, self.device,
                          self.soft_to_hard_fn, self.soft_to_hard_fn_kwargs, self._eval_loss_function)
        else:
            logger.info(
                "No dataset computed for validation on triggered dataset!")
            running_val_triggered_acc = None
            val_triggered_loss = None

        validation_stats = EpochValidationStatistics(running_val_clean_acc, val_clean_loss,
                                                     running_val_triggered_acc, val_triggered_loss)
        if num_val_batches_clean > 0:
            logger.info('{}\tTrain Epoch: {} \tCleanValLoss: {:.6f}\tCleanValAcc: {:.6f}'.format(
                pid, epoch_num, val_clean_loss, running_val_clean_acc))
        if num_val_batches_triggered > 0:
            logger.info('{}\tTrain Epoch: {} \tTriggeredValLoss: {:.6f}\tTriggeredValAcc: {:.6f}'.format(
                pid, epoch_num, val_triggered_loss, running_val_triggered_acc))

        if self.tb_writer:
            try:
                batch_num = int((epoch_num + 1) * num_batches)
                if num_val_batches_clean > 0:
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-clean-val-loss', val_clean_loss, global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-clean-val_acc', running_val_clean_acc, global_step=batch_num)
                if num_val_batches_triggered > 0:
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-triggered-val-loss', val_triggered_loss, global_step=batch_num)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                              '-triggered-val_acc', running_val_triggered_acc, global_step=batch_num)
            except:
                pass

        # update the lr-scheduler if necessary
        if self.lr_scheduler is not None:
            if self.optimizer_cfg.training_cfg.lr_scheduler_call_arg is None:
                self.lr_scheduler.step()
            elif self.optimizer_cfg.training_cfg.lr_scheduler_call_arg.lower() == 'val_acc':
                val_acc = validation_stats.get_val_acc()
                if val_acc is not None:
                    self.lr_scheduler.step(val_acc)
                else:
                    msg = "val_clean_acc not defined b/c validation dataset is not defined! Ignoring LR step!"
                    logger.warning(msg)
            elif self.optimizer_cfg.training_cfg.lr_scheduler_call_arg.lower() == 'val_loss':
                val_loss = validation_stats.get_val_loss()
                if val_loss is not None:
                    self.lr_scheduler.step(val_loss)
                else:
                    msg = "val_clean_loss not defined b/c validation dataset is not defined! Ignoring LR step!"
                    logger.warning(msg)
            else:
                msg = "Unknown mode for calling lr_scheduler!"
                logger.error(msg)
                raise ValueError(msg)

        return train_stats, validation_stats



from typing import Dict

import math
import numpy as np
from PIL import Image, ImageEnhance, ImageStat

import logging
logger = logging.getLogger(__name__)

"""
Module contains various utilities to modify the lighting of an image.
"""


def find_image_exposure(img_input: np.ndarray) -> Dict:
    """
    Gets various metrics related to the image exposure, and returns them in a dictionary
    :param img_input: the input image, for which exposure metrics should be calculated
    :return:
    """
    n_chan = img_input.shape[2]
    img = Image.fromarray(img_input.astype(np.uint8))
    im = Image.fromarray(img_input.astype(np.uint8)).convert('LA')
    stat = ImageStat.Stat(im)
    # Average pixel brighness
    avg = stat.mean[0]
    # RMS pixel brighness
    rms = stat.rms[0]
    stat2 = ImageStat.Stat(img)

    # Consider the number of channels
    # background may have RGB while traffic sign has RGBA
    if n_chan == 3:
        # Average pixels preceived brightness
        r, g, b = stat2.mean
        avg_perceived = math.sqrt(
            0.241 * (r ** 2) + 0.691 * (g ** 2) + 0.068 * (b ** 2))
        # RMS pixels perceived brightness
        r, g, b = stat2.rms
        rms_perceived = math.sqrt(
            0.241 * (r ** 2) + 0.691 * (g ** 2) + 0.068 * (b ** 2))
        # exposure_info = [avg, rms, avg_perceived, rms_perceived]
    else:
        # Average pixels preceived brightness
        r, g, b, a = stat2.mean
        avg_perceived = math.sqrt(
            0.241 * (r ** 2) + 0.691 * (g ** 2) + 0.068 * (b ** 2))
        # RMS pixels perceived brightness
        r, g, b, a = stat2.rms
        rms_perceived = math.sqrt(
            0.241 * (r ** 2) + 0.691 * (g ** 2) + 0.068 * (b ** 2))
        # exposure_info = [avg, rms, avg_perceived, rms_perceived]

    exposure_info = dict(avg=avg, rms=rms, perceived_avg=avg_perceived, perceived_rms=rms_perceived)
    logger.debug("Computed image exposure info: %s" % (str(exposure_info),))
    return exposure_info


def adjust_brightness_mmavg(fg_img: np.ndarray, bg_img: np.ndarray, avrg_ratio: float = 11.01) -> np.ndarray:
    """
    Adjust's brightness by minimizing the margin based on the average of the two channel brightness variations.

    :param fg_img: The foreground image
    :param bg_img: The background image
    :param avrg_ratio: the ratio of brightness to maintain between foreground and background images according to
            the average criterion
    :return: the brightness adjust image
    """
    peak = Image.fromarray(fg_img.astype(np.uint8)).convert('LA')
    stat = ImageStat.Stat(peak)
    avrg = stat.mean[0]
    peak2 = Image.fromarray(fg_img.astype(np.uint8)).convert('RGBA')
    enhancer = ImageEnhance.Brightness(peak2)
    background_exposures = find_image_exposure(bg_img)
    margin = abs(avrg - float(background_exposures['avg']))
    brightness_avrg = margin / avrg_ratio
    avrg_bright = enhancer.enhance(brightness_avrg)
    logger.debug("Enhanced brightness by %0.02f according to the minimize average channel margin method" %
                (brightness_avrg,))
    return np.array(avrg_bright)


def adjust_brightness_mmrms(fg_img: np.ndarray, bg_img: np.ndarray, rms_ratio: float = 8.3) -> np.ndarray:
    """
    Adjust's brightness by minimizing the margin based on the RMS of the two channel brightness variations.

    :param fg_img: The foreground image
    :param bg_img: The background image
    :param rms_ratio: the ratio of brightness to maintain between foreground and background images according to
            the RMS criterion
    :return: the brightness adjust image
    """
    peak = Image.fromarray(fg_img.astype(np.uint8)).convert('LA')
    stat = ImageStat.Stat(peak)
    rms = stat.rms[0]
    peak2 = Image.fromarray(fg_img.astype(np.uint8)).convert('RGBA')
    enhancer = ImageEnhance.Brightness(peak2)
    background_exposures = find_image_exposure(bg_img)
    margin = abs(rms - float(background_exposures['rms']))
    brightness_avrg = margin / rms_ratio
    rms_bright = enhancer.enhance(brightness_avrg)
    logger.debug("Enhanced brightness by %0.02f according to the minimize RMS channel margin method" %
                (brightness_avrg,))
    return np.array(rms_bright)


def adjust_brightness_mmpavg(fg_img: np.ndarray, bg_img: np.ndarray, percieved_avrg_ratio: float = 3.85) -> np.ndarray:
    """
    Adjust's brightness by minimizing the margin based on the perceived average of the two channel brightness variations
    REFERENCE FOR ALGORITHM USED: http://alienryderflex.com/hsp.html

    :param fg_img: The foreground image
    :param bg_img: The background image
    :param percieved_avrg_ratio: the ratio of brightness to maintain between foreground and background images according
            to the perceived average criterion
    :return: the brightness adjust image
    """
    peak2 = Image.fromarray(fg_img.astype(np.uint8)).convert('RGBA')
    stat2 = ImageStat.Stat(peak2)
    r, g, b, a = stat2.mean
    avrg_perceived = math.sqrt(
        0.241 * (r ** 2) + 0.691 * (g ** 2) + 0.068 * (b ** 2))

    enhancer = ImageEnhance.Brightness(peak2)
    background_exposures = find_image_exposure(bg_img)
    margin = abs(avrg_perceived - float(background_exposures['perceived_avg']))
    brightness_avrg = margin / percieved_avrg_ratio
    rms_bright = enhancer.enhance(brightness_avrg)
    logger.debug("Enhanced brightness by %0.02f according to the perceived average difference between channels method" %
                (brightness_avrg,))
    return np.array(rms_bright)


def adjust_brightness_mmprms(fg_img: np.ndarray, bg_img: np.ndarray, percieved_rms_ratio: float = 35.6) -> np.ndarray:
    """
    Adjust's brightness by minimizing the margin based on the perceived RMS of the two channel brightness variations
    REFERENCE FOR ALGORITHM USED: http://alienryderflex.com/hsp.html

    :param fg_img: The foreground image
    :param bg_img: The background image
    :param percieved_rms_ratio: the ratio of brightness to maintain between foreground and background images according
            to the perceived RMS criterion
    :return: the brightness adjust image
    """
    peak2 = Image.fromarray(fg_img.astype(np.uint8)).convert('RGBA')
    stat2 = ImageStat.Stat(peak2)
    r, g, b, a = stat2.mean
    avrg_perceived = math.sqrt(
        0.241 * (r ** 2) + 0.691 * (g ** 2) + 0.068 * (b ** 2))

    enhancer = ImageEnhance.Brightness(peak2)
    background_exposures = find_image_exposure(bg_img)
    margin = abs(avrg_perceived - float(background_exposures['perceived_rms']))
    brightness_avrg = margin / percieved_rms_ratio
    rms_bright = enhancer.enhance(brightness_avrg)
    logger.debug("Enhanced brightness by %0.02f according to the perceived RMS difference between channels method" %
                (brightness_avrg,))
    return np.array(rms_bright)



#!/usr/bin/env python3

import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

__author__ = 'Kiran Karra, Chace Ashcraft, Nat Kavaler, Michael Majurski, Taylor Kulp-McDowall'
__email__ = 'kiran.karra@jhuapl.edu,chace.ashcraft@jhuapl.edu,nathaniel.kavaler@jhaupl.edu,' \
            'michael.majurski@nist.gov,taylor.kulp-mcdowall@iarpa.gov'
__version__ = '0.2.19'
# Additional credit for software design attributed to:
#  Cash Costello: cash.costello@jhuapl.edu
#  Nathan Drenkow: nathan.drenkow@jhuapl.edu
#  Neil Fendley: neil.fendley@jhuapl.edu

on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    install_requires = []
else:
    install_requires = ['numpy>=1.19.2',
                        'pandas>=1.0.3',
                        'scikit-image>=0.17.2',
                        'joblib>=0.14.1',
                        'scipy>=1.5.3',
                        'pillow>=7.1.2',
                        'scikit-learn>=0.23.0',
                        'tqdm>=4.46.0',
                        'opencv-python>=4.2.0.34',
                        'torch==1.7.0',
                        'torchvision==0.8.0',
                        'torchtext==0.8.0',
                        'blend_modes',
                        'spacy>=2.2.0',
                        'cloudpickle>=1.4.1',
                        'Wand>=0.5.9',
                        'nltk>=3.5',
                        'pyllist>=0.3',
                        'albumentations',
                        'blend_modes',
                        'advertorch',
                        'nltk',
                        'pyllist',
                        'transformers==4.5.1',  # TODO: this requires a rust compiler being installed on OSX ...
                        'kaggle',
                        'tb-nightly',
                        'future'
                        ]

setuptools.setup(
    name='trojai',
    version=__version__,

    description='TrojAI model and dataset generation library',
    long_description=long_description,
    long_description_content_type="text/markdown",

    url = 'https://github.com/trojai/trojai',

    author=__author__,
    author_email=__email__,

    license='Apache License 2.0',

    python_requires='>=3',
    packages=['trojai', 'trojai.datagen', 'trojai.modelgen', 'trojai.modelgen.architectures'],

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='deep-learning trojan adversarial',

    # Needed for albumentations install
    dependency_links=['git+https://bitbucket.xrcs.jhuapl.edu/scm/troj/'
                      'albumentations.git@reproducible_randomness_for_augmentations#egg=albumentations'],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html

    install_requires=install_requires,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
          'test': ['nose==1.3.7', 'coverage==5.0.3', 'mock']
    },

    scripts=['scripts/datagen/mnist.py',
             'scripts/datagen/mnist_utils.py',
             'scripts/datagen/mnist_badnets.py',
             'scripts/datagen/mnist_badnets2.py',
             'scripts/datagen/mnist_badnets_one_class_trigger.py',
             'scripts/modelgen/gen_and_train_mnist.py',
             'scripts/modelgen/gen_and_train_mnist_sequential.py'
             ],

    zip_safe=False
)



import logging

import cv2
import numpy as np
from numpy.random import RandomState

from .image_entity import ImageEntity, GenericImageEntity
from .transform_interface import Transform

logger = logging.getLogger(__name__)

"""
Module contains various classes that relate to size transformations of input objects
"""


class Resize(Transform):
    """
    Resizes an Entity
    """
    def __init__(self, new_size: tuple = (200, 200), interpolation: int = cv2.INTER_CUBIC) -> None:
        """
        Initialize the resizer object
        :param new_size: a tuple of the size in pixes for x and y dimensions
        :param interpolation: the interpolation method to resize the input Entity
        """
        self.new_size = new_size
        self.interpolation = interpolation

    def do(self, img_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the resizing
        :param img_obj: The input object to be resized according the specified configuration
        :param random_state_obj: ignored
        :return: The resized object
        """
        img_out = cv2.resize(img_obj.get_data(), self.new_size, interpolation=self.interpolation)
        mask_out = cv2.resize(img_obj.get_mask().astype(np.float32), self.new_size,
                              interpolation=self.interpolation).astype(bool)
        logger.debug("Resized image of shape=%s to shape=%s using %d interpolation" %
                    (str(img_obj.get_data().shape), str(self.new_size), self.interpolation))
        return GenericImageEntity(img_out, mask_out)


class RandomResize(Transform):
    """
    Resizes an Entity
    """
    def __init__(self, new_size_minimum: tuple = (200, 200), new_size_maximum: tuple = (300, 300), interpolation: int = cv2.INTER_CUBIC) -> None:
        """
        Initialize the resizer object
        :param new_size_minimum: a tuple of the minimum size in pixes for x and y dimensions
        :param new_size_maximum: a tuple of the maximum size in pixes for x and y dimensions
        :param interpolation: the interpolation method to resize the input Entity
        """
        self.new_size_minimum = new_size_minimum
        self.new_size_maximum = new_size_maximum
        self.interpolation = interpolation

    def do(self, img_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the resizing
        :param img_obj: The input object to be resized according the specified configuration
        :param random_state_obj: ignored
        :return: The resized object
        """
        # select a new size from within the range of calid new sizes
        y = random_state_obj.randint(self.new_size_minimum[0], self.new_size_maximum[0])
        x = random_state_obj.randint(self.new_size_minimum[1], self.new_size_maximum[1])
        new_size = (y, x)

        img_out = cv2.resize(img_obj.get_data(), new_size, interpolation=self.interpolation)
        mask_out = cv2.resize(img_obj.get_mask().astype(np.float32), new_size,
                              interpolation=self.interpolation).astype(bool)
        logger.debug("Resized image of shape=%s to shape=%s using %d interpolation" %
                    (str(img_obj.get_data().shape), str(new_size), self.interpolation))
        return GenericImageEntity(img_out, mask_out)


class RandomPadToSize(Transform):
    """
    Resizes an Entity
    """
    def __init__(self, new_size: tuple = (200, 200), mode: str = 'constant', pad_value: int = 0) -> None:
        """
        Initialize the resizer object
        :param new_size: a tuple of the size in pixes for x and y dimensions
        :param mode: what type of padding to use, supports numpy.pad options
        :param pad_value: the value to use when padding
        """
        self.new_size = new_size
        self.mode = mode
        self.pad_value = pad_value

    def do(self, img_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the resizing
        :param img_obj: The input object to be resized according the specified configuration
        :param random_state_obj: ignored
        :return: The resized object
        """

        input_shape = img_obj.get_data().shape

        if self.new_size[0] < input_shape[0] or self.new_size[1] < input_shape[1]:
            raise RuntimeError('Invalid pad new_size {} smaller than input image size {}'.format(self.new_size, input_shape))

        total_pad_y = self.new_size[0] - input_shape[0]
        total_pad_x = self.new_size[1] - input_shape[1]

        pre_pad_value_y = random_state_obj.randint(0, total_pad_y)
        pre_pad_value_x = random_state_obj.randint(0, total_pad_x)
        pad_values = (pre_pad_value_y, total_pad_y - pre_pad_value_y, pre_pad_value_x, total_pad_x - pre_pad_value_x)

        return Pad(pad_values, self.mode, self.pad_value).do(img_obj, random_state_obj)


class Pad(Transform):
    """
    Resizes an Entity
    """
    def __init__(self, pad_amounts: tuple = (0, 0, 0, 0), mode: str = 'constant', pad_value: int = 0) -> None:
        """
        Initialize the resizer object
        :param pad_amounts: a tuple of the pixel count o add to each side (y_pre, y_post, x_pre, x_post)
        :param mode: what type of padding to use, supports numpy.pad options
        :param pad_value: the value to use when padding
        """
        self.pad_amounts = pad_amounts
        self.mode = mode
        self.pad_value = pad_value

    def do(self, img_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the resizing
        :param img_obj: The input object to be resized according the specified configuration
        :param random_state_obj: ignored
        :return: The resized object
        """

        if np.count_nonzero(np.asarray(self.pad_amounts)) == 0:
            return img_obj

        img = img_obj.get_data()
        msk = img_obj.get_mask()
        kwargs = {}
        if self.mode == 'constant':
            kwargs = {'constant_values': self.pad_value}

        if len(img.shape) == 2:
                img_out = np.pad(img, pad_width=((self.pad_amounts[0], self.pad_amounts[1]), (self.pad_amounts[2], self.pad_amounts[3])), mode=self.mode, **kwargs)
        elif len(img.shape) == 3:
            img_out = np.pad(img, pad_width=((self.pad_amounts[0], self.pad_amounts[1]), (self.pad_amounts[2], self.pad_amounts[3]), (0, 0)), mode=self.mode, **kwargs)
        else:
            raise RuntimeError('Unexpected image shape: {}'.format(img.shape))

        if len(msk.shape) == 2:
            mask_out = np.pad(msk, pad_width=((self.pad_amounts[0], self.pad_amounts[1]), (self.pad_amounts[2], self.pad_amounts[3])), mode=self.mode, **kwargs)
        elif len(msk.shape) == 3:
            mask_out = np.pad(msk, pad_width=((self.pad_amounts[0], self.pad_amounts[1]), (self.pad_amounts[2], self.pad_amounts[3]), (0, 0)), mode=self.mode, **kwargs)
        else:
            raise RuntimeError('Unexpected mask shape: {}'.format(msk.shape))

        logger.debug("Padded image of shape=%s to shape=%s" %
                    (str(img_obj.get_data().shape), str(img.shape)))
        return GenericImageEntity(img_out, mask_out)


class RandomSubCrop(Transform):
    """
    Resizes an Entity
    """
    def __init__(self, new_size: tuple = (200, 200)) -> None:
        """
        Initialize the crop object
        :param new_size: a tuple of the size in pixels for x and y dimensions
        """
        self.new_size = new_size

    def do(self, img_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the resizing
        :param img_obj: The input object to be cropped according the specified configuration
        :param random_state_obj: ignored
        :return: The cropped object
        """
        img = img_obj.get_data()
        msk = img_obj.get_mask()

        if self.new_size[0] > img.shape[0]:
            raise RuntimeError('Invalid subcrop size: requested height {} is larger than source image height {}'.format(self.new_size[0], img.shape[0]))
        if self.new_size[1] > img.shape[1]:
            raise RuntimeError('Invalid subcrop size: requested width {} is larger than source image width {}'.format(self.new_size[1], img.shape[1]))

        if self.new_size[0] == img.shape[0] and self.new_size[1] == img.shape[1]:
            return img_obj

        y_st = np.random.randint(0, img.shape[0] - self.new_size[0])
        x_st = np.random.randint(0, img.shape[1] - self.new_size[1])

        img_out = img[y_st:y_st + self.new_size[0], x_st:x_st + self.new_size[1]]
        mask_out = msk[y_st:y_st + self.new_size[0], x_st:x_st + self.new_size[1]]

        logger.debug("Cropped source image size {} to output size {}".format(img_obj.get_data().shape, img_out.shape))
        return GenericImageEntity(img_out, mask_out)



import cv2
import numpy as np

from trojai.datagen.image_entity import ImageEntity


class FlatIconDotComPng(ImageEntity):
    """
    Defines a png icon for a trigger.
    """
    def __init__(self, trigger_fpath, mode='graffiti', trigger_color=None, postit_bg_color=None, size=None):
        """
        Initializes a trigger from a png file
        :param trigger_fpath: filepath to the png image defining the trigger.
        :param mode: trigger mode.
        :param trigger_color: trigger color RGB.
        :param postit_bg_color: trigger background color RBG.
        :param size: trigger target size.
        """

        if postit_bg_color is None:
            postit_bg_color = [0, 0, 0]   # default black background
        if trigger_color is None:
            trigger_color = [255, 255, 255]    # default white trigger

        self.data = cv2.imread(trigger_fpath, cv2.IMREAD_UNCHANGED)
        if size is not None:
            self.data = cv2.resize(self.data, dsize=size, interpolation=cv2.INTER_NEAREST)

        if mode.lower() == 'graffiti':
            self.mask = (self.data[:,:,3] > 0).astype(bool)
            for c in range(3):
                self.data[:, :, c] = trigger_color[c]
        elif mode.lower() == 'postit':
            self.mask = np.ones((self.data.shape[0], self.data.shape[1]), dtype=bool)
            data_new = np.zeros((self.data.shape[0], self.data.shape[1], 3), dtype=np.uint8)
            ident_mat = np.ones((self.data.shape[0], self.data.shape[1]), dtype=np.uint8)
            np.putmask(data_new[:, :, 0], self.data[:, :, 3].astype(bool), trigger_color[0] * ident_mat)
            np.putmask(data_new[:, :, 0], ~self.data[:, :, 3].astype(bool), postit_bg_color[0] * ident_mat)
            np.putmask(data_new[:, :, 1], self.data[:, :, 3].astype(bool), trigger_color[1] * ident_mat)
            np.putmask(data_new[:, :, 1], ~self.data[:, :, 3].astype(bool), postit_bg_color[1] * ident_mat)
            np.putmask(data_new[:, :, 2], self.data[:, :, 3].astype(bool), trigger_color[2] * ident_mat)
            np.putmask(data_new[:, :, 2], ~self.data[:, :, 3].astype(bool), postit_bg_color[2] * ident_mat)
            self.data = cv2.cvtColor(data_new, cv2.COLOR_RGB2RGBA)

    def get_data(self) -> np.ndarray:
        return self.data

    def get_mask(self) -> np.ndarray:
        return self.mask



import collections.abc
import json
import logging
from typing import Union, Sequence
import csv
import numpy as np

logger = logging.getLogger(__name__)

"""
Contains classes necessary for collecting statistics on the model during training
"""


class BatchStatistics:
    """
    Represents the statistics collected from training a batch
    NOTE: this is currently unused!
    """
    def __init__(self, batch_num: int,
                 batch_train_accuracy: float,
                 batch_train_loss: float):
        """
        :param batch_num: (int) batch number of collected statistics
        :param batch_train_accuracy: (float) training set accuracy for this batch
        :param batch_train_loss: (float) training loss for this batch
        """
        self.batch_num = batch_num
        self.batch_train_accuracy = batch_train_accuracy
        self.batch_train_loss = batch_train_loss

    def get_batch_num(self):
        return self.batch_num

    def get_batch_train_acc(self):
        return self.batch_train_accuracy

    def get_batch_train_loss(self):
        return self.batch_train_loss

    def set_batch_train_acc(self, acc):
        if 0 <= acc <= 100:
            self.batch_train_accuracy = acc
        else:
            msg = "Batch training accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_batch_train_loss(self, loss):
        self.batch_train_loss = loss


class EpochTrainStatistics:
    """
    Defines the training statistics for one epoch of training
    """
    def __init__(self, train_acc: float, train_loss: float):
        self.train_acc = train_acc
        self.train_loss = train_loss

        self.validate()

    def validate(self):
        if not isinstance(self.train_acc, float):
            msg = "train_acc must be a float, got type {}".format(type(self.train_acc))
            logger.error(msg)
            raise ValueError(msg)

        if not isinstance(self.train_loss, float):
            msg = "train_loss must be a float, got type {}".format(type(self.train_loss))
            logger.error(msg)
            raise ValueError(msg)

    def get_train_acc(self):
        return self.train_acc

    def get_train_loss(self):
        return self.train_loss


class EpochValidationStatistics:
    """
    Defines the validation statistics for one epoch of training
    """
    def __init__(self, val_clean_acc, val_clean_loss, val_triggered_acc, val_triggered_loss):
        self.val_clean_acc = val_clean_acc
        self.val_clean_loss = val_clean_loss
        self.val_triggered_acc = val_triggered_acc
        self.val_triggered_loss = val_triggered_loss

        self.validate()

    def validate(self):
        if self.val_clean_acc is not None and not isinstance(self.val_clean_acc, float):
            msg = "val_clean_acc must be a float, got type {}".format(type(self.val_clean_acc))
            logger.error(msg)
            raise ValueError(msg)

        if self.val_clean_loss is not None and not isinstance(self.val_clean_loss, float):
            msg = "val_clean_loss must be a float, got type {}".format(type(self.val_clean_loss))
            logger.error(msg)
            raise ValueError(msg)

        if self.val_triggered_acc is not None and not isinstance(self.val_triggered_acc, float):
            msg = "val_triggered_acc must be a float, got type {}".format(type(self.val_triggered_acc))
            logger.error(msg)
            raise ValueError(msg)

        if self.val_triggered_loss is not None and not isinstance(self.val_triggered_loss, float):
            msg = "val_triggered_loss must be a float, got type {}".format(type(self.val_triggered_loss))
            logger.error(msg)
            raise ValueError(msg)

    def get_val_clean_acc(self):
        return self.val_clean_acc

    def get_val_clean_loss(self):
        return self.val_clean_loss

    def get_val_triggered_acc(self):
        return self.val_triggered_acc

    def get_val_triggered_loss(self):
        return self.val_triggered_loss

    def get_val_loss(self):
        if self.get_val_triggered_loss() is not None and self.get_val_clean_loss() is not None:
            return self.get_val_triggered_loss() + self.get_val_clean_loss()
        elif self.get_val_triggered_loss() is None and self.get_val_clean_loss() is not None:
            return self.get_val_clean_loss()
        elif self.get_val_triggered_loss() is not None and self.get_val_clean_loss() is None:
            return self.get_val_triggered_loss()
        else:
            return None

    def get_val_acc(self):
        if self.get_val_triggered_acc() is not None and self.get_val_clean_acc() is not None:
            return (self.get_val_triggered_acc() + self.get_val_clean_acc())/2.
        elif self.get_val_triggered_acc() is None and self.get_val_clean_acc() is not None:
            return self.get_val_clean_acc()
        elif self.get_val_triggered_acc() is not None and self.get_val_clean_acc() is None:
            return self.get_val_triggered_acc()
        else:
            return None

    def __repr__(self):
        val_loss = self.get_val_loss()
        val_acc = self.get_val_acc()
        val_loss = val_loss if val_loss is not None else -999
        val_acc = val_acc if val_acc is not None else -999

        return '(%0.04f, %0.04f)' % (val_loss, val_acc)


class EpochStatistics:
    """
    Contains the statistics computed for an Epoch
    """
    def __init__(self, epoch_num, training_stats=None, validation_stats=None, batch_training_stats=None):
        self.epoch_num = epoch_num
        if not batch_training_stats:
            self.batch_training_stats = []
        self.epoch_training_stats = training_stats
        self.epoch_validation_stats = validation_stats

        self.validate()

    def add_batch(self, batches: Union[BatchStatistics, Sequence[BatchStatistics]]):
        if isinstance(batches, collections.abc.Sequence):
            self.batch_training_stats.extend(batches)
        else:
            self.batch_training_stats.append(batches)

    def get_batch_stats(self):
        return self.batch_training_stats

    def validate(self):
        if not isinstance(self.batch_training_stats, collections.abc.Sequence):
            msg = "batch_training_stats must be None or a list of BatchTrainingStats objects! " \
                  "Got {}".format(self.batch_training_stats)
            logger.error(msg)
            raise ValueError(msg)
        if self.epoch_training_stats and not isinstance(self.epoch_training_stats, EpochTrainStatistics):
            msg = "training_stats must be None or of type: EpochTrainStatistics!, got type " \
                  "{}".format(type(self.epoch_training_stats))
            logger.error(msg)
            raise ValueError(msg)
        if self.epoch_validation_stats and not isinstance(self.epoch_validation_stats, EpochValidationStatistics):
            msg = "validation_stats must be None or of type: EpochValidationStatistics! Instead got type " \
                  "{}".format(type(self.epoch_validation_stats))
            logger.error(msg)
            raise ValueError(msg)

    def get_epoch_num(self):
        return self.epoch_num

    def get_epoch_training_stats(self):
        return self.epoch_training_stats

    def get_epoch_validation_stats(self):
        return self.epoch_validation_stats


class TrainingRunStatistics:
    """
    Contains the statistics computed for an entire training run, a sequence of epochs
    TODO:
     [ ] - have another function which returns detailed statistics per epoch in an easily serialized manner
    """
    def __init__(self):
        self.stats_per_epoch_list = []

        self.num_epochs_trained_per_optimizer = []

        self.final_train_acc = 0.
        self.final_train_loss = 0.
        self.final_combined_val_acc = 0.
        self.final_combined_val_loss = 0.
        self.final_clean_val_acc = 0.
        self.final_clean_val_loss = 0.
        self.final_triggered_val_acc = 0.
        self.final_triggered_val_loss = 0.
        self.final_clean_data_test_acc = 0.
        self.final_clean_data_n_total = 0
        self.final_triggered_data_test_acc = None
        self.final_triggered_data_n_total = None
        self.final_clean_data_triggered_labels_test_acc = None
        self.final_clean_data_triggered_labels_n_total = None
        self.final_optimizer_num_epochs_trained = 0
        self.final_optimizer_best_epoch_val = -1

    def add_epoch(self, epoch_stats: Union[EpochStatistics, Sequence[EpochStatistics]]):
        if isinstance(epoch_stats, collections.abc.Sequence):
            self.stats_per_epoch_list.extend(epoch_stats)
        else:
            self.stats_per_epoch_list.append(epoch_stats)

    def add_num_epochs_trained(self, num_epochs):
        self.num_epochs_trained_per_optimizer.append(num_epochs)

    def add_best_epoch_val(self, best_epoch):
        self.final_optimizer_best_epoch_val = best_epoch

    def get_epochs_stats(self):
        return self.stats_per_epoch_list

    def autopopulate_final_summary_stats(self):
        """
        Uses the information from the final epoch's final batch to auto-populate the following statistics:
            final_train_acc
            final_train_loss
            final_val_acc
            final_val_loss
        """
        final_epoch_training_stats = self.stats_per_epoch_list[self.final_optimizer_best_epoch_val]

        self.set_final_train_acc(final_epoch_training_stats.get_epoch_training_stats().get_train_acc())
        self.set_final_train_loss(final_epoch_training_stats.get_epoch_training_stats().get_train_loss())
        if final_epoch_training_stats.get_epoch_validation_stats():
            self.set_final_val_combined_acc(final_epoch_training_stats.get_epoch_validation_stats().get_val_acc())
            self.set_final_val_combined_loss(final_epoch_training_stats.get_epoch_validation_stats().get_val_loss())

            self.set_final_val_clean_acc(final_epoch_training_stats.get_epoch_validation_stats().get_val_clean_acc())
            self.set_final_val_clean_loss(final_epoch_training_stats.get_epoch_validation_stats().get_val_clean_loss())
            self.set_final_val_triggered_acc(final_epoch_training_stats.get_epoch_validation_stats().get_val_triggered_acc())
            self.set_final_val_triggered_loss(final_epoch_training_stats.get_epoch_validation_stats().get_val_triggered_loss())

        self.final_optimizer_num_epochs_trained = self.num_epochs_trained_per_optimizer[-1]

    def set_final_train_acc(self, acc):
        if 0 <= acc <= 100:
            self.final_train_acc = acc
        else:
            msg = "Final Training accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_train_loss(self, loss):
        self.final_train_loss = loss

    def set_final_val_combined_acc(self, acc):
        if acc is None or 0 <= acc <= 100:  # allow for None in case validation metrics are not computed
            self.final_combined_val_acc = acc
        else:
            msg = "Final validation accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_val_combined_loss(self, loss):
        self.final_combined_val_loss = loss

    def set_final_val_clean_acc(self, acc):
        self.final_clean_val_acc = acc

    def set_final_val_triggered_acc(self, acc):
        self.final_triggered_val_acc = acc

    def set_final_val_clean_loss(self, loss):
        self.final_clean_val_loss = loss

    def set_final_val_triggered_loss(self, loss):
        self.final_triggered_val_loss = loss

    def set_final_clean_data_test_acc(self, acc):
        if 0 <= acc <= 100:
            self.final_clean_data_test_acc = acc
        else:
            msg = "Final clean data test accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_triggered_data_test_acc(self, acc):
        # we allow None in order to indicate that triggered data wasn't present in this dataset
        if acc is None or 0 <= acc <= 100:
            self.final_triggered_data_test_acc = acc
        else:
            msg = "Final triggered data test accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_clean_data_triggered_label_test_acc(self, acc):
        if acc is None or 0 <= acc <= 100:
            self.final_clean_data_triggered_labels_test_acc = acc
        else:
            msg = "Final clean data test accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_clean_data_n_total(self, n):
        self.final_clean_data_n_total = n

    def set_final_triggered_data_n_total(self, n):
        self.final_triggered_data_n_total = n

    def set_final_clean_data_triggered_label_n(self, n):
        self.final_clean_data_triggered_labels_n_total = n

    def get_summary(self):
        """
        Returns a dictionary of the summary statistics from the training run
        """
        summary_dict = dict()
        summary_dict['final_train_acc'] = self.final_train_acc
        summary_dict['final_train_loss'] = self.final_train_loss
        summary_dict['final_combined_val_acc'] = self.final_combined_val_acc
        summary_dict['final_combined_val_loss'] = self.final_combined_val_loss
        summary_dict['final_clean_val_acc'] = self.final_clean_val_acc
        summary_dict['final_clean_val_loss'] = self.final_clean_val_loss
        summary_dict['final_triggered_val_acc'] = self.final_triggered_val_acc
        summary_dict['final_triggered_val_loss'] = self.final_triggered_val_loss
        summary_dict['final_clean_data_test_acc'] = self.final_clean_data_test_acc
        summary_dict['final_triggered_data_test_acc'] = self.final_triggered_data_test_acc
        summary_dict['final_clean_data_n_total'] = self.final_clean_data_n_total
        summary_dict['final_triggered_data_n_total'] = self.final_triggered_data_n_total
        summary_dict['clean_test_triggered_label_accuracy'] = self.final_clean_data_triggered_labels_test_acc
        summary_dict['clean_test_triggered_label_n_total'] = self.final_clean_data_triggered_labels_n_total
        summary_dict['final_optimizer_num_epochs_trained'] = self.num_epochs_trained_per_optimizer

        return summary_dict

    def save_summary_to_json(self, json_fname: str) -> None:
        """
        Saves the training summary to a JSON file
        """
        summary_dict = self.get_summary()
        # write it to json
        with open(json_fname, 'w') as fp:
            json.dump(summary_dict, fp)
        logger.info("Wrote summary statistics: %s to %s" % (str(summary_dict), json_fname))

    def save_detailed_stats_to_disk(self, fname: str) -> None:
        """
        Saves all batch statistics for every epoch as a CSV file

        :param fname: filename to save the detailed information to
        :return: None
        """
        keys = ['epoch_number', 'train_acc', 'train_loss', 'combined_val_acc', 'combined_val_loss',
                'clean_val_acc', 'clean_val_loss', 'triggered_val_acc', 'triggered_val_loss']
        with open(fname, 'w') as output_file:
            # write header first
            dict_writer = csv.DictWriter(output_file, keys)
            dict_writer.writeheader()
            for ii, e in enumerate(self.stats_per_epoch_list):
                # TODO: we ignore batch_statistics for now, we may want to add this in in the future
                epoch_training_stats = e.get_epoch_training_stats()
                epoch_val_stats = e.get_epoch_validation_stats()
                combined_val_acc = None
                combined_val_loss = None
                clean_val_acc = None
                clean_val_loss = None
                triggered_val_acc = None
                triggered_val_loss = None
                if epoch_val_stats is not None:
                    combined_val_acc = epoch_val_stats.get_val_acc()
                    combined_val_loss = epoch_val_stats.get_val_loss()
                    clean_val_acc = epoch_val_stats.get_val_clean_acc()
                    clean_val_loss = epoch_val_stats.get_val_clean_loss()
                    triggered_val_acc = epoch_val_stats.get_val_triggered_acc()
                    triggered_val_loss = epoch_val_stats.get_val_triggered_loss()

                dict_writer.writerow(dict(epoch_number=e.get_epoch_num(),
                                          train_acc=epoch_training_stats.get_train_acc(),
                                          train_loss=epoch_training_stats.get_train_loss(),
                                          combined_val_acc=combined_val_acc,
                                          combined_val_loss=combined_val_loss,
                                          clean_val_acc=clean_val_acc,
                                          clean_val_loss=clean_val_loss,
                                          triggered_val_acc=triggered_val_acc,
                                          triggered_val_loss=triggered_val_loss))

            logger.info("Wrote detailed statistics to %s" % (fname,))



import logging
from typing import Sequence, Dict

import skimage.transform
import numpy as np
from numpy.random import RandomState
import cv2

from .image_entity import GenericImageEntity, ImageEntity
from .transform_interface import ImageTransform

logger = logging.getLogger(__name__)

"""
Module defines several affine transforms using various libraries to perform  the actual transformation operation 
specified.
"""

class UniformScaleXForm(ImageTransform):
    """Implements a uniform scale of a specified amount to an Entity

    """
    def __init__(self, scale_factor: float = 1, kwargs: dict = None) -> None:
        """
        Create a scaler object
        :param scale_factor: the scaling amount
        :param kwargs: any keyword arguments to pass to skimge.transform.rescale
        """
        self.scale_factor = scale_factor
        if kwargs is None:
            self.kwargs = {}
        else:
            self.kwargs = kwargs

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Performs the scaling on an input Entity using skimage.transform.rescale
        :param input_obj: the input object to be scaled
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        mask = input_obj.get_mask()

        logger.debug("Applying %0.02f scaling of image" % (self.scale_factor,))
        img_rescaled = skimage.transform.rescale(img, self.scale_factor, **self.kwargs)
        logger.debug("Applying %0.02f scaling of mask" % (self.scale_factor,))
        mask_rescaled = skimage.transform.rescale(mask, self.scale_factor, **self.kwargs)

        return GenericImageEntity(img_rescaled, mask_rescaled)


valid_predefined_xform_strs = [
    # NOTE: these have a large effect
    'east',
    'north-west',
    'shrink-1',
    'shrink-2',

    # NOTE: these have a medium effect
    'left-tilt-forward',
    'right-tilt-forward',
    'west',

    # NOTE: these have a small effect
    'forward-distortion-1',
    'forward-distortion-2',
    'forward-distortion-3',
    'forward-distortion-4',
    'forward-distortion-5',
    'forward-distortion-6',
    'forward-distortion-7',
    'forward-distortion-8',
    'forward-distortion-9',
    'forward-distortion-10',
    'forward-distortion-11',

    # NOTE: these have no effect
    'forward',
]


def get_predefined_perspective_xform_matrix(xform_str: str, rows: int, cols: int) -> np.ndarray:
    """
    Returns an affine transform matrix for a string specification of a
    perspective transformation
    :param xform_str: a string specification of the perspective to transform
           the object into.
    :param rows: the number of rows of the image to be transformed to the
           specified perspective
    :param cols: the number of cols of the image to be transformed to the
           specified perspective
    :return: a numpy array of shape (2,3) which specifies the affine
             transformation.

    See:https://docs.opencv.org/2.4/modules/imgproc/doc/geometric_transformations.html?highlight=getaffinetransform
    for more information
    """
    xform_str_lower = xform_str.lower()
    if xform_str_lower not in valid_predefined_xform_strs:
        raise ValueError("Unknown perspective transformation string!")

    if xform_str_lower == 'forward':
        return np.asarray([[1, 0, 0], [0, 1, 0]], dtype=np.float32)
    elif xform_str_lower == 'east':
        pts1 = np.float32([[cols / 10, rows / 10], [cols / 2, rows / 10], [cols / 10, rows / 2]])
        pts2 = np.float32([[cols / 5, rows / 5], [cols / 2, rows / 8], [cols / 5, rows / 1.8]])
    elif xform_str_lower == 'north-west':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 4.5 / 5, rows / 5], [cols / 2, rows / 8], [cols * 4.5 / 5, rows / 1.8]])
    elif xform_str_lower == 'left-tilt-forward':
        pts1 = np.float32([[cols / 10, rows / 10], [cols / 2, rows / 10], [cols / 10, rows / 2]])
        pts2 = np.float32([[cols / 12, rows / 6], [cols / 2.1, rows / 8], [cols / 10, rows / 1.8]])
    elif xform_str_lower == 'right-tilt-forward':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 10 / 12, rows / 6], [cols / 2.2, rows / 8], [cols * 8.4 / 10, rows / 1.8]])
    elif xform_str_lower == 'west':
        pts1 = np.float32([[cols / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols / 9.95, rows / 10], [cols / 2.05, rows / 9.95], [cols * 9 / 10, rows / 2.05]])
    elif xform_str_lower == 'forward-distortion-1':
        pts1 = np.float32([[cols / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols / 9.8, rows / 9.8], [cols / 2, rows / 9.8], [cols * 8.8 / 10, rows / 2.05]])
    elif xform_str_lower == 'forward-distortion-2':
        pts1 = np.float32([[cols / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols / 11, rows / 10], [cols / 2.1, rows / 10], [cols * 8.5 / 10, rows / 1.95]])
    elif xform_str_lower == 'forward-distortion-3':
        pts1 = np.float32([[cols / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols / 11, rows / 11], [cols / 2.1, rows / 10], [cols * 10 / 11, rows / 1.95]])
    elif xform_str_lower == 'forward-distortion-4':
        pts1 = np.float32([[cols * 9.5 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 9.35 / 10, rows / 9.99],
                           [cols / 2.05, rows / 9.95], [cols * 9.05 / 10, rows / 2.03]])
    elif xform_str_lower == 'forward-distortion-5':
        pts1 = np.float32([[cols * 9.5 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 9.65 / 10, rows / 9.95], [cols / 1.95, rows / 9.95], [cols * 9.1 / 10, rows / 2.02]])
    elif xform_str_lower == 'forward-distortion-6':
        pts1 = np.float32([[cols * 9.25 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 9.55 / 10, rows / 9.85], [cols / 1.9, rows / 10], [cols * 9.3 / 10, rows / 2.04]])
    elif xform_str_lower == 'forward-distortion-7':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 8.85 / 10, rows / 9.3], [cols / 1.9, rows / 10.5], [cols * 8.8 / 10, rows / 2.11]])
    elif xform_str_lower == 'forward-distortion-8':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 8.75 / 10, rows / 9.1], [cols / 1.95, rows / 8], [cols * 8.5 / 10, rows / 2.05]])
    elif xform_str_lower == 'forward-distortion-9':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 8.75 / 10, rows / 9.1], [cols / 1.95, rows / 9], [cols * 8.5 / 10, rows / 2.2]])
    elif xform_str_lower == 'forward-distortion-10':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 8.75 / 10, rows / 8], [cols / 1.95, rows / 8], [cols * 8.75 / 10, rows / 2]])
    elif xform_str_lower == 'forward-distortion-11':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 8.8 / 10, rows / 7], [cols / 1.95, rows / 7], [cols * 8.8 / 10, rows / 2]])
    elif xform_str_lower == 'shrink-1':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 8 / 10, rows / 10], [cols * 1.34 / 3, rows / 10.5], [cols * 8.24 / 10, rows / 2.5]])
    elif xform_str_lower == 'shrink-2':
        pts1 = np.float32([[cols * 9 / 10, rows / 10], [cols / 2, rows / 10], [cols * 9 / 10, rows / 2]])
        pts2 = np.float32([[cols * 8.5 / 10, rows * 3.1 / 10], [cols / 2, rows * 3 / 10],
                          [cols * 8.44 / 10, rows * 1.55 / 2.5]])
    else:
        raise ValueError("Unknown perspective transformation string!")
    return cv2.getAffineTransform(pts1, pts2)


class PerspectiveXForm(ImageTransform):
    """Shifts the perspective of an input Entity

    """
    def __init__(self, xform_matrix) -> None:
        """
        Creates a Perspective shifter object
        :param xform_matrix: can be either a string specification of a perspective shift, where valid strings are
               defined in the list: affine_xforms.valid_predefined_xform_strs, or it can be a matrix of shape (2,3).
        """
        # input validation
        if isinstance(xform_matrix, str):
            self.xform_M = xform_matrix
        elif isinstance(xform_matrix, np.ndarray) and xform_matrix.shape == (2, 3):
            self.xform_M = xform_matrix
        else:
            raise ValueError("Unknown M input, must be either an allowed string or a matrix of shape (2,3)!")

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Performs the perspective shift on the input Entity.
        :param input_obj: the Entity to be transformed according to the specified perspective shift in the constructor.
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        mask = input_obj.get_mask()
        i_rows, i_cols, i_chans = img.shape
        if isinstance(self.xform_M, str):
            xform_matrix = get_predefined_perspective_xform_matrix(self.xform_M, i_rows, i_cols)
        else:
            xform_matrix = self.xform_M

        logger.debug("Applying cv2.warpAffine to image with matrix:" + str(xform_matrix))
        img_xform = cv2.warpAffine(img, xform_matrix, (i_cols, i_rows))
        logger.debug("Applying cv2.warpAffine to mask with matrix:" + str(xform_matrix))
        msk_xform = cv2.warpAffine(mask.astype(np.float32), xform_matrix, (i_cols, i_rows)).astype(bool)

        return GenericImageEntity(img_xform, msk_xform)


class RandomPerspectiveXForm(ImageTransform):
    """Randomly shifts perspective of input Entity in available perspectives.

    """
    def __init__(self, perspectives: Sequence[str] = None) -> None:
        """
        Creates a random perspective shifter Transform object, which uniformly samples the available perspectives in
        AffineXForms.valid_predefined_xform_strs

        # TODO: add support for non-uniform sampling of perspective transformations
        """
        if perspectives is None:
            self.perspective_possibilities = valid_predefined_xform_strs
        else:
            for perspective in perspectives:
                if perspective not in valid_predefined_xform_strs:
                    msg = perspective + " is not in the valid list of transforms"
                    logger.error(msg)
                    raise ValueError(msg)
            self.perspective_possibilities = perspectives

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Samples from the possible perspectives according to the sampler
        specification and then applies that perspective to the input object
        :param input_obj: Entity to be randomly perspective shifted
        :param random_state_obj: allows for reprodcible sampling of random perspectives
        :return: the transformed Entity
        """
        # pick a perspective transformation
        chosen_xform = random_state_obj.choice(self.perspective_possibilities)

        logger.debug("Sampled perspective %s from RandomState" % (chosen_xform,))
        xformer = PerspectiveXForm(chosen_xform)

        return xformer.do(input_obj, random_state_obj)


class RotateXForm(ImageTransform):
    """Implements a rotation of an Entity by a specified angle amount.

    """
    def __init__(self, angle: int = 90, args: tuple = (), kwargs: dict = None) -> None:
        """
        Creates a Rotator Transform object
        :param angle: The degree amount to rotate (in degrees, not radians!)
        :param args: any additional arguments to pass to skimage.transform.rotate
        :param kwargs: any keyword arguments to pass to skimage.transform.rotate
        """
        self.rotation_angle = angle
        self.args = args
        if kwargs is None:
            self.kwargs = {'preserve_range': True}
        else:
            if 'preserve_range' in kwargs and not kwargs['preserve_range']:
                msg = "preserve_range cannot be set to False!"
                logger.error(msg)
                raise ValueError(msg)
            self.kwargs = kwargs

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Performs the rotation specified by the RotateXForm object on an input
        :param input_obj: The Entity to be rotated
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        mask = input_obj.get_mask()

        logger.debug("Applying %0.02f rotation to image via skimage.transform.rotate" % (self.rotation_angle,))
        img_rotated = skimage.transform.rotate(img, self.rotation_angle, *self.args, **self.kwargs).astype(img.dtype)
        logger.debug("Applying %0.02f rotation to mask via skimage.transform.rotate" % (self.rotation_angle,))
        mask_rotated = skimage.transform.rotate(mask, self.rotation_angle, *self.args, **self.kwargs)
        mask_rotated = np.logical_not(np.isclose(mask_rotated, np.zeros(mask.shape), atol=.0001))

        return GenericImageEntity(img_rotated, mask_rotated)


class RandomRotateXForm(ImageTransform):
    """Implements a rotation of a random amount of degrees.

    """
    def __init__(self, angle_choices: Sequence[float] = None, angle_sampler_prob: Sequence[float] = None,
                 rotator_kwargs: Dict = None) -> None:
        """
        Creates a random rotator Transform object
        :param angle_choices: An Sequence object of floats which represent the possible angles
                              from which the sampler can choose from
        :param angle_sampler_kwargs: any keyword arguments to pass to the sampler
        """
        if angle_choices is None:
            self.angle_choices = [0, 90, 180, 270]
        else:
            self.angle_choices = angle_choices
        if rotator_kwargs is None:
            self.rotator_kwargs = {'preserve_range': True}
        else:
            self.rotator_kwargs = rotator_kwargs
        self.angle_sampler_prob = angle_sampler_prob

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Samples from the possible angles according to the sampler specification and then applies that
        rotation to the input object
        :param input_obj: Entity to be randomly rotated
        :param random_state_obj: a random state used to maintain reproducibility through transformations
        :return: the transformed Entity
        """
        rotation_angle = random_state_obj.choice(self.angle_choices, p=self.angle_sampler_prob)

        logger.debug("Sampled %0.02f rotation from RandomState" % (rotation_angle,))
        rotator = RotateXForm(rotation_angle, kwargs=self.rotator_kwargs)

        return rotator.do(input_obj, random_state_obj)



import logging

import torch
import torch.nn as nn

logger = logging.getLogger(__name__)

from transformers import BertTokenizer, BertModel


class EmbeddingLSTM(nn.Module):
    """
    Defines an LSTM model that can be used for text classification
    from: https://github.com/bentrevett/pytorch-sentiment-analysis/blob/master/2%20-%20Upgraded%20Sentiment
            %20Analysis.ipynb
    """
    def __init__(self, vocab_size, embedding_dim, hidden_dim, output_dim, n_layers, bidirectional, dropout, pad_idx):
        super().__init__()

        logger.info("Creating EmbeddingLSTM model with Embedding [" + str(vocab_size) +
                    " x " + str(embedding_dim) + "]. pad_idx=" + str(pad_idx))
        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.n_layers = n_layers
        self.bidirectional = bidirectional
        self.dropout = dropout
        self.pad_idx = pad_idx

        self.embedding = nn.Embedding(vocab_size, embedding_dim, padding_idx=pad_idx)
        self.rnn = nn.LSTM(embedding_dim,
                           hidden_dim,
                           num_layers=n_layers,
                           bidirectional=bidirectional,
                           dropout=dropout)
        self.fc = nn.Linear(hidden_dim * 2, output_dim)
        self.dropout = nn.Dropout(dropout)

        self.packed_padded_sequences = True

    def forward(self, text, text_lengths):
        # text.shape = [sentence len, batch size]
        # embedded.shape = [sentence len, batch size, emb dim]
        embedded = self.dropout(self.embedding(text))

        # pack sequence
        packed_embedded = nn.utils.rnn.pack_padded_sequence(embedded, text_lengths)
        self.rnn.flatten_parameters()
        packed_output, (hidden, cell) = self.rnn(packed_embedded)
        # hidden.shape = [num layers * num directions, batch size, hid dim]
        # cell.shape = [num layers * num directions, batch size, hid dim]

        # concat the final forward (hidden[-2,:,:]) and backward (hidden[-1,:,:]) hidden layers and apply dropout
        hidden = self.dropout(torch.cat((hidden[-2, :, :], hidden[-1, :, :]), dim=1))
        # hidden.shape  = [batch size, hid dim * num directions]

        return self.fc(hidden)


class EmbeddingGRU(nn.Module):
    """
    Defines an GRU model that can be used for text classification
    from: https://github.com/bentrevett/pytorch-sentiment-analysis/blob/master/2%20-%20Upgraded%20Sentiment
            %20Analysis.ipynb
    """
    def __init__(self, vocab_size, embedding_dim, hidden_dim, output_dim, n_layers, bidirectional, dropout, pad_idx):
        super().__init__()

        logger.info("Creating EmbeddingGRU model with Embedding [" + str(vocab_size) +
                    " x " + str(embedding_dim) + "]. pad_idx=" + str(pad_idx))
        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.n_layers = n_layers
        self.bidirectional = bidirectional
        self.dropout = dropout
        self.pad_idx = pad_idx

        self.embedding = nn.Embedding(vocab_size, embedding_dim, padding_idx=pad_idx)
        self.rnn = nn.GRU(embedding_dim,
                          hidden_dim,
                          num_layers=n_layers,
                          bidirectional=bidirectional,
                          batch_first=True,
                          dropout=0 if n_layers < 2 else dropout)
        self.fc = nn.Linear(hidden_dim * 2, output_dim)
        self.dropout = nn.Dropout(dropout)

        self.packed_padded_sequences = True

    def forward(self, text, text_lengths):
        # text.shape = [sentence len, batch size]
        # embedded.shape = [sentence len, batch size, emb dim]
        embedded = self.dropout(self.embedding(text))

        # pack sequence
        packed_embedded = nn.utils.rnn.pack_padded_sequence(embedded, text_lengths)
        packed_output, (hidden, cell) = self.rnn(packed_embedded)
        # hidden.shape = [num layers * num directions, batch size, hid dim]
        # cell.shape = [num layers * num directions, batch size, hid dim]

        # concat the final forward (hidden[-2,:,:]) and backward (hidden[-1,:,:]) hidden layers and apply dropout
        hidden = self.dropout(torch.cat((hidden[-2, :, :], hidden[-1, :, :]), dim=1))
        # hidden.shape  = [batch size, hid dim * num directions]

        return self.fc(hidden)


class XFormerGRU(nn.Module):
    """
    From: https://github.com/bentrevett/pytorch-sentiment-analysis/blob/master/6%20-%20Transformers%20for%20Sentiment%20Analysis.ipynb
    """
    def __init__(self, xformer, hidden_dim, output_dim, n_layers, bidirectional, dropout, freeze_bert=True):
        super().__init__()
        self.xformer = xformer
        embedding_dim = xformer.config.to_dict()['hidden_size']

        self.rnn = nn.GRU(embedding_dim,
                          hidden_dim,
                          num_layers=n_layers,
                          bidirectional=bidirectional,
                          batch_first=True,
                          dropout=0 if n_layers < 2 else dropout)

        self.out = nn.Linear(hidden_dim * 2 if bidirectional else hidden_dim, output_dim)
        self.dropout = nn.Dropout(dropout)

        self.packed_padded_sequences = False

        if freeze_bert:
            for name, param in self.named_parameters():
                if name.startswith('xformer'):
                    param.requires_grad = False

    def forward(self, text):

        # text = [batch size, sent len]
        with torch.no_grad():
            embedded = self.xformer(text)[0]

        # embedded = [batch size, sent len, emb dim]
        _, hidden = self.rnn(embedded)

        # hidden = [n layers * n directions, batch size, emb dim]
        if self.rnn.bidirectional:
            hidden = self.dropout(torch.cat((hidden[-2, :, :], hidden[-1, :, :]), dim=1))
        else:
            hidden = self.dropout(hidden[-1, :, :])

        # hidden = [batch size, hid dim]
        output = self.out(hidden)
        # output = [batch size, out dim]

        return output


def pretrained_bert_gru_sentiment_model(hidden_dim, output_dim, n_layers, bidirectional, dropout, freeze_bert=True):
    """

    :return:
    """
    bert = BertModel.from_pretrained('bert-base-uncased')

    # build the model
    model = XFormerGRU(bert, hidden_dim, output_dim, n_layers, bidirectional, dropout, freeze_bert=freeze_bert)

    return model


def bert_gru_sentiment_model(hidden_dim, output_dim, n_layers, bidirectional, dropout, **kwargs):
    pretrained_bert = BertModel.from_pretrained('bert-base-uncased')
    bert_config = copy.copy(pretrained_bert.config)
    # NOTE: I don't think a deep copy is needed here because the BertConfig object doesn't
    #  have any compound objects.
    #  See here: https://huggingface.co/transformers/_modules/transformers/configuration_bert.html#BertConfig
    pretrained_bert = None  # try to get the garbage collector to clean this up from memory

    # TODO: update this to be more efficient (i.e. get the config directly from the pretrained function rather than
    #  instantiating an object and then pulling hte ocnfig from that) - for now, we are just testing things out
    #  I think this is the config: https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-uncased-config.json
    raw_bert = BertModel(bert_config)
    # build the model
    model = XFormerGRU(raw_bert, hidden_dim, output_dim, n_layers, bidirectional, dropout, freeze_bert=False)

    return model


def bert_tokenizer():
    # The vocabulary for this tokenizer is here:
    #  https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-uncased-vocab.txt
    bert_tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
    return bert_tokenizer